from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal
import pytest

setup_logging("debug")


@MPITest([4])
def test_bad_input(comm):

    source = UniformCatalog(3e-4, BoxSize=256, seed=42, comm=comm)
    source['gal_type'] = 1

    # cannot rank by missing column
    with pytest.raises(ValueError):
        r = CylindricalGroups(source, 'missing', 10, 10, periodic=True, flat_sky_los=[0,0,1])

    # bad flat_sky_los vector
    with pytest.raises(ValueError):
        r = CylindricalGroups(source, 'gal_type', 10, 10, periodic=True, flat_sky_los=[0,0,0,1])
    with pytest.raises(ValueError):
        r = CylindricalGroups(source, 'gal_type', 10, 10, periodic=True, flat_sky_los=[1,1,1]) # not a unit vector

    # missing BoxSize!
    del source.attrs['BoxSize']
    with pytest.raises(ValueError):
        r = CylindricalGroups(source, 'gal_type', 10, 10, periodic=True, flat_sky_los=[0,0,1])

    # this should work though -- we specified BoxSize
    r = CylindricalGroups(source, 'gal_type', 10, 10, BoxSize=256.0, periodic=False)

    # warn if periodic and no line-of-sight
    with pytest.warns(UserWarning):
        r = CylindricalGroups(source, 'gal_type', 10, 10, BoxSize=256.0, periodic=True)


@MPITest([4])
def test_periodic_cgm(comm):

    source = UniformCatalog(3e-4, BoxSize=256, seed=42, comm=comm)

    # add mass
    logmass = source.rng.uniform(12, 15)
    source['halo_mvir'] = 10**(logmass)

    # add fake galaxy types
    gal_type = numpy.empty(len(source))
    gal_type[logmass<14.5] = 0
    gal_type[logmass>14.5] = 1
    source['gal_type'] = gal_type

    # run the algorithm
    rankby = ['halo_mvir', 'gal_type']
    rpar = 10.0
    rperp = 10.0
    r = CylindricalGroups(source, rpar=rpar, rperp=rperp, rankby=rankby, periodic=True, flat_sky_los=[0,0,1])

    # data for direct CGM
    pos = numpy.concatenate(comm.allgather(source['Position']), axis=0)
    mass = numpy.concatenate(comm.allgather(source['halo_mvir']), axis=0)
    gal_type = numpy.concatenate(comm.allgather(source['gal_type']), axis=0)

    # direct results
    kws = {'periodic':True, 'BoxSize':source.attrs['BoxSize']}
    N_cgm, cgm_gal_type, cen_id = direct_cgm(pos, mass, gal_type, rperp, rpar, **kws)

    # gather and compare
    N_cgm2 = numpy.concatenate(comm.allgather(r.groups['num_cgm_sats']), axis=0)
    cen_id2 = numpy.concatenate(comm.allgather(r.groups['cgm_haloid']), axis=0)
    cgm_gal_type2 = numpy.concatenate(comm.allgather(r.groups['cgm_type']), axis=0)

    assert_array_equal(N_cgm, N_cgm2)
    assert_array_equal(cen_id, cen_id2)
    assert_array_equal(cgm_gal_type, cgm_gal_type2)


@MPITest([1, 4])
def test_nonperiodic_cgm(comm):

    source = UniformCatalog(3e-4, BoxSize=256, seed=42, comm=comm)

    # add mass
    logmass = source.rng.uniform(12, 15)
    source['halo_mvir'] = 10**(logmass)

    # add fake galaxy types
    gal_type = numpy.empty(len(source))
    gal_type[logmass<14.5] = 0
    gal_type[logmass>14.5] = 1
    source['gal_type'] = gal_type

    # run the algorithm
    rankby = ['halo_mvir', 'gal_type']
    rpar = 10.0
    rperp = 10.0
    r = CylindricalGroups(source, rpar=rpar, rperp=rperp, rankby=rankby, periodic=False, flat_sky_los=[0,0,1])

    # data for direct CGM
    pos = numpy.concatenate(comm.allgather(source['Position']), axis=0)
    mass = numpy.concatenate(comm.allgather(source['halo_mvir']), axis=0)
    gal_type = numpy.concatenate(comm.allgather(source['gal_type']), axis=0)

    # direct results
    N_cgm, cgm_gal_type, cen_id = direct_cgm(pos, mass, gal_type, rperp, rpar, periodic=False)

    # gather and compare
    N_cgm2 = numpy.concatenate(comm.allgather(r.groups['num_cgm_sats']), axis=0)
    cen_id2 = numpy.concatenate(comm.allgather(r.groups['cgm_haloid']), axis=0)
    cgm_gal_type2 = numpy.concatenate(comm.allgather(r.groups['cgm_type']), axis=0)

    assert_array_equal(N_cgm, N_cgm2)
    assert_array_equal(cen_id, cen_id2)
    assert_array_equal(cgm_gal_type, cgm_gal_type2)


def direct_cgm(pos, mass, gal_type, rperp, rpar, periodic=False, BoxSize=None):
    """
    Given the position of particles, and the mass and galaxy type data
    to sort by, return the non-periodic CGM results via directly comparing
    each object

    Notes
    -----
    * The function is not collective; all data must be passed to this function
    * The line-of-sight is assumed to be [0,0,1]
    """
    N = len(pos)

    # initialize output arrays
    N_cgm = numpy.zeros(N, dtype='i8')
    cgm_gal_type = numpy.zeros(N, dtype='u1')
    cen_id = numpy.zeros(N, dtype='i8') - 1

    # sort
    dtype = numpy.dtype([
            ('origind', 'u4'),
            ('mass', mass.dtype),
            ('gal_type', gal_type.dtype),
            ('pos', (pos.dtype.str, 3))
            ])
    data = numpy.empty(N, dtype=dtype)
    data['mass'] = mass
    data['gal_type'] = gal_type
    data['pos'] = pos
    data['origind'] = numpy.arange(len(data), dtype='u4')

    # sort pos1 by mass and then gal type
    data.sort(order=['mass', 'gal_type'])
    data = data[::-1]

    ii = 0
    while len(data):

        dr = data['pos'][0] - data['pos'][1:]
        if periodic:
            for axis, col in enumerate(dr.T):
                col[col > BoxSize[axis]*0.5] -= BoxSize[axis]
                col[col <= -BoxSize[axis]*0.5] += BoxSize[axis]
        dr2 = (dr**2).sum(axis=-1)
        rflat_sky_los = dr[:,-1]
        rsky2 = numpy.abs(dr2 - rflat_sky_los ** 2)

        valid = (abs(rflat_sky_los) <= rpar)&(rsky2 <= rperp**2)
        this_cenid = data['origind'][0]
        N_cgm[this_cenid] = valid.sum()
        cgm_gal_type[data['origind'][1:][valid]] = 1
        cen_id[data['origind'][1:][valid]] = ii
        cen_id[data['origind'][0]] = ii

        # delete central
        data = numpy.delete(data, 0, axis=0)

        #delete satellites
        data = numpy.delete(data, numpy.where(valid==True)[0], axis=0)

        ii += 1

    return N_cgm, cgm_gal_type, cen_id



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_allclose

# debug logging
setup_logging("debug")

@MPITest([1, 4])
def test_fof(comm):
    cosmo = cosmology.Planck15

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=128., Nmesh=32, seed=42, comm=comm)

    # compute P(k,mu) and multipoles
    fof = FOF(source, linking_length=0.2, nmin=20)
    source['Density'] = KDDensity(source, margin=1).density

    # save the halos
    peaks = fof.find_features(peakcolumn='Density')
    peaks.save("FOF-%d" % comm.size, ['CMPosition', 'CMVelocity', 'Length', 'PeakPosition', 'PeakVelocity'])

@MPITest([1, 4])
def test_fof_parallel_no_merge(comm):
    from pmesh.pm import ParticleMesh
    pm = ParticleMesh(BoxSize=[8, 8, 8], Nmesh=[8, 8, 8], comm=comm)
    Q = pm.generate_uniform_particle_grid()
    cat = ArrayCatalog({'Position' : Q}, BoxSize=pm.BoxSize, Nmesh=pm.Nmesh, comm=comm)

    fof = FOF(cat, linking_length=0.9, nmin=0)
    
    labels = numpy.concatenate(comm.allgather((fof.labels)), axis=0)
    # one particle per group
    assert max(labels) == cat.csize - 1

@MPITest([1, 4])
def test_fof_parallel_merge(comm):
    from pmesh.pm import ParticleMesh
    pm = ParticleMesh(BoxSize=[8, 8, 8], Nmesh=[8, 8, 8], comm=comm)
    Q = pm.generate_uniform_particle_grid(shift=0)
    Q1 = Q.copy()
    Q1[:] += 0.01
    Q2 = Q.copy()
    Q2[:] -= 0.01
    Q3 = Q.copy()
    Q3[:] += 0.02
    cat = ArrayCatalog({'Position' : 
            numpy.concatenate([Q, Q1, Q2, Q3], axis=0)}, BoxSize=pm.BoxSize, Nmesh=pm.Nmesh, comm=comm)

    fof = FOF(cat, linking_length=0.011 * 3 ** 0.5, nmin=0, absolute=True)

    labels = numpy.concatenate(comm.allgather((fof.labels)), axis=0)
    assert max(labels) == pm.Nmesh.prod() - 1
    assert all(numpy.bincount(labels) == 4)

@MPITest([1, 4])
def test_fof_nonperiodic(comm):
    cosmo = cosmology.Planck15

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=128., Nmesh=32, seed=42, comm=comm)

    source['Density'] = KDDensity(source, margin=1).density

    del source.attrs['BoxSize'] # no boxsize 

    # shift left
    source['Position'] -= 100.0
    fof = FOF(source, linking_length=0.2, nmin=20, periodic=False, absolute=True)
    peaks1 = fof.find_features(peakcolumn='Density')

    # shift right
    source['Position'] += 200.0
    fof = FOF(source, linking_length=0.2, nmin=20, periodic=False, absolute=True)
    peaks2 = fof.find_features(peakcolumn='Density')

    assert_allclose(peaks1['CMPosition'] + 200.0, peaks2['CMPosition'], rtol=1e-6)
    assert_allclose(peaks1['CMVelocity'], peaks2['CMVelocity'], rtol=1e-6)
    assert_allclose(peaks1['PeakPosition'] + 200.0, peaks2['PeakPosition'], rtol=1e-6)
    assert_allclose(peaks1['PeakVelocity'], peaks2['PeakVelocity'], rtol=1e-6)



from runtests.mpi import MPITest
from nbodykit import setup_logging
from nbodykit.mpirng import MPIRandomState
from numpy.testing import assert_array_equal
import numpy
from mpi4py import MPI
import os
import pytest

setup_logging("debug")

@MPITest([4])
def test_mpirng_large_chunk(comm):
    rng = MPIRandomState(comm, seed=1234, size=1, chunksize=10)

    local = rng.uniform()
    all = numpy.concatenate(comm.allgather(local), axis=0)

    rng1 = MPIRandomState(MPI.COMM_SELF, seed=1234, size=rng.csize, chunksize=rng.chunksize)

    correct = rng1.uniform()

    assert_array_equal(all, correct)

@MPITest([4])
def test_mpirng_small_chunk(comm):
    rng = MPIRandomState(comm, seed=1234, size=10, chunksize=3)

    local = rng.uniform()
    all = numpy.concatenate(comm.allgather(local), axis=0)

    rng1 = MPIRandomState(MPI.COMM_SELF, seed=1234, size=rng.csize, chunksize=rng.chunksize)

    correct = rng1.uniform()

    assert_array_equal(all, correct)

@MPITest([4])
def test_mpirng_unique(comm):
    rng = MPIRandomState(comm, seed=1234, size=10, chunksize=3)

    local1 = rng.uniform()
    local2 = rng.uniform()

    # it shouldn't be the same!
    assert (local1 != local2).any()

@MPITest([4])
def test_mpirng_args(comm):
    rng = MPIRandomState(comm, seed=1234, size=10, chunksize=3)

    local = rng.uniform(low=numpy.ones(rng.size) * 0.5)
    all = numpy.concatenate(comm.allgather(local), axis=0)

    rng1 = MPIRandomState(MPI.COMM_SELF, seed=1234, size=rng.csize, chunksize=rng.chunksize)

    correct = rng1.uniform(low=0.5)

    assert_array_equal(all, correct)

@MPITest([4])
def test_mpirng_itemshape(comm):
    rng = MPIRandomState(comm, seed=1234, size=10, chunksize=3)

    local = rng.uniform(low=numpy.ones(rng.size)[:, None] * 0.5, itemshape=(3,))
    all = numpy.concatenate(comm.allgather(local), axis=0)

    rng1 = MPIRandomState(MPI.COMM_SELF, seed=1234, size=rng.csize, chunksize=rng.chunksize)

    correct = rng1.uniform(low=0.5, itemshape=(3,))

    assert_array_equal(all, correct)

@MPITest([4])
def test_mpirng_poisson(comm):
    rng = MPIRandomState(comm, seed=1234, size=10, chunksize=3)

    local = rng.poisson(lam=numpy.ones(rng.size)[:, None] * 0.5, itemshape=(3,))
    all = numpy.concatenate(comm.allgather(local), axis=0)

    rng1 = MPIRandomState(MPI.COMM_SELF, seed=1234, size=rng.csize, chunksize=rng.chunksize)

    correct = rng1.poisson(lam=0.5, itemshape=(3,))

    assert_array_equal(all, correct)




from .array import ArrayCatalog
from nbodykit import CurrentMPIComm, transform
from nbodykit.utils import GatherArray, ScatterArray
from nbodykit.base.catalog import CatalogSourceBase, CatalogSource, column

import numpy
import logging

class HaloCatalog(CatalogSource):
    """
    A CatalogSource of objects that represent halos, which can be populated
    using analytic models from :mod:`halotools`.

    Parameters
    ----------
    source : CatalogSource
        the source holding the particles to be interpreted as halos
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`
        the cosmology instance;
    redshift : float
        the redshift of the halo catalog
    mdef : str, optional
        string specifying mass definition, used for computing default
        halo radii and concentration; should be 'vir' or 'XXXc' or
        'XXXm' where 'XXX' is an int specifying the overdensity
    mass : str, optional
        the column name specifying the mass of each halo
    position : str, optional
        the column name specifying the position of each halo
    velocity : str, optional
        the column name specifying the velocity of each halo
    """
    logger = logging.getLogger("HaloCatalog")

    def __init__(self, source, cosmo, redshift, mdef='vir',
                 mass='Mass', position='Position', velocity='Velocity'):

        # make sure all of the columns are there
        required = ['mass', 'position', 'velocity']
        for name, col in zip(required, [mass, position, velocity]):
            if col is None:
                raise ValueError("the %s column cannot be None in HaloCatalog" %name)
            if col not in source:
                raise ValueError("input source is missing the %s column; '%s' does not exist" %(name, col))

        if not isinstance(source, CatalogSourceBase):
            raise TypeError("input source to HalotoolsCatalog should be a CatalogSource")

        comm = source.comm
        self._source = source

        self.cosmo = cosmo

        # get the attrs from the source
        self.attrs.update(source.attrs)

        # and save the parameters
        self.attrs['redshift'] = redshift
        self.attrs['cosmo']    = dict(self.cosmo)
        self.attrs['mass']     = mass
        self.attrs['velocity'] = velocity
        self.attrs['position'] = position
        self.attrs['mdef']     = mdef

        # names of the mass and radius fields, based on mass def
        self.attrs['halo_mass_key'] = 'halo_m' + mdef
        self.attrs['halo_radius_key'] = 'halo_r' + mdef

        # the size
        self._size = self._source.size

        # init the base class
        CatalogSource.__init__(self, comm=comm)

    @column
    def Mass(self):
        """
        The halo mass column, assumed to be in units of :math:`M_\odot/h`.
        """
        return self._source[self.attrs['mass']]

    @column
    def Position(self):
        """
        The halo position column, assumed to be in units of :math:`\mathrm{Mpc}/h`.
        """
        return self._source[self.attrs['position']]

    @column
    def Velocity(self):
        """
        The halo velocity column, assumed to be in units of km/s.
        """
        return self._source[self.attrs['velocity']]

    @column
    def VelocityOffset(self):
        """
        The redshift-space distance offset due to the velocity in units of
        distance. The assumed units are :math:`\mathrm{Mpc}/h`.

        This multiplies ``Velocity`` by :math:`1 / (a 100 E(z)) = 1 / (a H(z)/h)`.
        """
        z = self.attrs['redshift']
        rsd_factor = (1+z) / (100*self.cosmo.efunc(z))
        return self['Velocity'] * rsd_factor

    @column
    def Concentration(self):
        """
        The halo concentration, computed using :func:`nbodykit.transform.HaloConcentration`.

        This uses the analytic formulas for concentration from
        `Dutton and Maccio 2014 <https://arxiv.org/abs/1402.7073>`_.

        Users can override this column to implement custom mass-concentration
        relations.
        """
        z = self.attrs['redshift']
        mdef = self.attrs['mdef']
        return transform.HaloConcentration(self['Mass'], self.cosmo, z, mdef=mdef)

    @column
    def Radius(self):
        """
        The halo radius, computed using :func:`nbodykit.transform.HaloRadius`.

        Assumed units of :math:`\mathrm{Mpc}/h`.
        """
        z = self.attrs['redshift']
        mdef = self.attrs['mdef']
        return transform.HaloRadius(self['Mass'], self.cosmo, z, mdef=mdef)

    def to_halotools(self, BoxSize=None):
        """
        Return the HaloCatalog as a
        :class:`halotools.sim_manager.UserSuppliedHaloCatalog`.

        The Halotools catalog only holds the local data, although halos are
        labeled via the ``halo_id`` column using the global index.

        Parameters
        ----------
        BoxSize : float, array_like, optional
            the size of the box; must be supplied if 'BoxSize' is not in
            the :attr:`attrs` dict

        Returns
        -------
        cat : :class:`halotools.sim_manager.UserSuppliedHaloCatalog`
            the Halotools halo catalog, storing the local halo data
        """
        from halotools.sim_manager import UserSuppliedHaloCatalog

        # make sure we have at least one halo
        if self.csize == 0:
            raise ValueError("cannot create halotools UserSuppliedHaloCatalog; catalog is empty")

        # make sure we have a BoxSize
        if BoxSize is None:
            BoxSize = self.attrs.get('BoxSize', None)
        if BoxSize is None:
            raise ValueError("please specify a 'BoxSize' to convert to a halotools catalog")

        # compute the columns
        cols = ['Position', 'Velocity', 'Mass', 'Radius', 'Concentration']
        cols = self.compute(*[self[col] for col in cols])
        Position, Velocity, Mass, Radius, Concen = [col for col in cols]

        # halo mass and radius keys
        mkey = self.attrs['halo_mass_key']
        rkey = self.attrs['halo_radius_key']

        # global halo ids (across all ranks)
        halo_id = self.Index.compute()

        # data columns
        kws                  = {}
        kws['halo_x']        = Position[:,0]
        kws['halo_y']        = Position[:,1]
        kws['halo_z']        = Position[:,2]
        kws['halo_vx']       = Velocity[:,0]
        kws['halo_vy']       = Velocity[:,1]
        kws['halo_vz']       = Velocity[:,2]
        kws[mkey]            = Mass
        kws[rkey]            = Radius
        kws['halo_nfw_conc'] = Concen
        kws['halo_id']       = halo_id
        kws['halo_hostid']    = halo_id
        kws['halo_upid']     = numpy.zeros(len(Position)) - 1
        kws['halo_local_id'] = numpy.arange(0, self.size, dtype='i8')

        # add metadata too
        kws['cosmology']     = self.cosmo
        kws['redshift']      = self.attrs['redshift']
        kws['Lbox']          = BoxSize
        kws['particle_mass'] = self.attrs.get('particle_mass', 1.0)
        kws['mdef']          = self.attrs['mdef']

        return UserSuppliedHaloCatalog(**kws)

    def populate(self, model, BoxSize=None, seed=None, **params):
        """
        Populate the HaloCatalog using a :mod:`halotools` model.

        The model can be a built-in model from :mod:`nbodykit.hod` (which
        will be converted to a Halotools model) or directly a Halotools model
        instance.

        This assumes that this is the first time this catalog has been
        populated with the input model. To re-populate using the same
        model (but different parameters), call the :func:`repopulate`
        function of the returned :class:`PopulatedHaloCatalog`.

        Parameters
        ----------
        model : :class:`nbodykit.hod.HODModel` or halotools model object
            the model instance to use to populate; model types from
            :mod:`nbodykit.hod` will automatically be converted
        BoxSize : float, 3-vector, optional
            the box size of the catalog; this must be supplied if 'BoxSize'
            is not in :attr:`attrs`
        seed : int, optional
            the random seed to use when populating the mock
        **params :
            key/value pairs specifying the model parameters to use

        Returns
        -------
        cat : :class:`PopulatedHaloCatalog`
            the catalog object storing information about the populated objects

        Examples
        --------
        Initialize a demo halo catalog:

        >>> from nbodykit.tutorials import DemoHaloCatalog
        >>> cat = DemoHaloCatalog('bolshoi', 'rockstar', 0.5)

        Populate with the built-in Zheng07 model:

        >>> from nbodykit.hod import Zheng07Model
        >>> galcat = cat.populate(Zheng07Model, seed=42)

        And then re-populate galaxy catalog with new parameters:

        >>> galcat.repopulate(alpha=0.9, logMmin=13.5, seed=42)
        """
        from nbodykit.hod import HODModel
        from halotools.empirical_models import ModelFactory
        from halotools.sim_manager import UserSuppliedHaloCatalog

        # handle builtin model types
        if isinstance(model, (type, HODModel)) and issubclass(model, HODModel):
            model = model.to_halotools(self.cosmo, self.attrs['redshift'],
                                        self.attrs['mdef'], concentration_key='halo_nfw_conc')

        # check model type
        if not isinstance(model, ModelFactory):
            raise TypeError("model for populating mocks should be a Halotools ModelFactory")

        # make halotools catalog
        halocat = self.to_halotools(BoxSize=BoxSize)

        # cache the model so we have option to call repopulate later
        self.model = model

        # return the populated catalog
        return _populate_mock(self, model, seed=seed, halocat=halocat, **params)


class PopulatedHaloCatalog(ArrayCatalog):
    """
    A CatalogSource to represent a set of objects populated into a
    :class:`HaloCatalog`.

    .. note::
        Users should not access this class directly, but rather, call
        :func:`HaloCatalog.populate` to generate a :class:`PopulatedHaloCatalog`.

    Parameters
    ----------
    data : structured numpy.ndarray
        the data of the populated objects
    model :
        the Halotools model instance
    cosmo : :class:`nbodykit.cosmology.cosmology.Cosmology`
        the cosmology instance
    """
    @CurrentMPIComm.enable
    def __init__(self, data, model, cosmo, comm=None):
        ArrayCatalog.__init__(self, data, comm=comm)
        self.model = model
        self.cosmo = cosmo

    def repopulate(self, seed=None, **params):
        """
        Re-populate the catalog in-place, using the specified ``seed``
        or model parameters.

        This re-uses the model that was last used to create this catalog.
        It is faster than :func:`HaloCatalog.populate` as it avoids
        initialization steps. It is intended to be used when looping over
        different parameter sets, e.g., when performing parameter optimization.

        .. note::
            This operation is performed in-place.

        Parameters
        ----------
        seed : int, optional
            the random seed to use when populating the mock
        **params :
            key/value pairs specifying the model parameters to use
        """
        _populate_mock(self, self.model, seed=seed, inplace=True, **params)


def _populate_mock(cat, model, seed=None, halocat=None, inplace=False, **params):
    """
    Internal function to perform the mock population on a HaloCatalog, given
    a :mod:`halotools` model.

    The implementation is not massively parallel. The data is gathered to
    the root rank, mock population is performed, and then the data is
    re-scattered evenly across ranks.
    """
    # verify input params
    valid = sorted(model.param_dict)
    missing = set(params) - set(valid)
    if len(missing):
        raise ValueError("invalid halo model parameter names: %s" % str(missing))

    # update the model parameters
    model.param_dict.update(params)

    # set the seed randomly if it is None
    if seed is None:
        seed = numpy.random.randint(0, 4294967295)

    # use uncorrelated seed per rank
    rng = numpy.random.RandomState(seed=seed)
    seed1 = rng.randint(0, 4294967295, size=cat.comm.size)[cat.comm.rank]

    # the types of galaxies we are populating
    gal_types = getattr(model, 'gal_types', [])

    # re-populate the mock (without halo catalog pre-processing)
    kws = {'seed':seed1, 'Num_ptcl_requirement':0, 'halo_mass_column_key':cat.attrs['halo_mass_key']}
    if hasattr(model, 'mock'):
        model.mock.populate(**kws)
    # populating model for the first time (initialization costs)
    else:
        if halocat is None:
            raise ValueError("halocat cannot be None if we are populating for the first time")
        model.populate_mock(halocat=halocat, **kws)

    # enumerate gal types as integers
    # NOTE: necessary to avoid "O" type columns
    _enum_gal_types(model.mock.galaxy_table, gal_types)

    # crash if any object dtypes
    # NOTE: we cannot use GatherArray/ScatterArray on objects
    data = _test_for_objects(model.mock.galaxy_table).as_array()

    # re-initialize with new source
    if inplace:
        PopulatedHaloCatalog.__init__(cat, data, model, cat.cosmo, comm=cat.comm)
        galcat = cat
    else:
        galcat = PopulatedHaloCatalog(data, model, cat.cosmo, comm=cat.comm)

    # crash with no particles!
    if galcat.csize == 0:
        raise ValueError("no particles in catalog after populating halo catalog")

    # add Position, Velocity
    galcat['Position'] = transform.StackColumns(galcat['x'], galcat['y'], galcat['z'])
    galcat['Velocity'] = transform.StackColumns(galcat['vx'], galcat['vy'], galcat['vz'])

    # add VelocityOffset
    z = cat.attrs['redshift']
    rsd_factor = (1+z) / (100.*cat.cosmo.efunc(z))
    galcat['VelocityOffset'] = galcat['Velocity'] * rsd_factor

    # add meta-data
    galcat.attrs.update(cat.attrs)
    galcat.attrs.update(model.param_dict)
    galcat.attrs['seed'] = seed
    galcat.attrs['gal_types'] = {t:i for i,t in enumerate(gal_types)}

    # propagate total number of halos for logging
    Nhalos = galcat.comm.allreduce(len(galcat.model.mock.halo_table))

    # and log some info
    _log_populated_stats(galcat, Nhalos)

    return galcat

def _log_populated_stats(cat, Nhalos):
    """
    Internal function to log statistics of a populated catalog. It logs
    information about satellite fraction, mass distribution, and total
    number.

    This is called each time that :func:`_populate_mock` is called.
    """
    # compute the satellite fraction
    fsat = None
    if 'gal_type' in cat:
        gal_types = cat.attrs['gal_types']
        if 'satellites' in gal_types:
            Nsats = cat.comm.allreduce((cat['gal_type'] == gal_types['satellites']).sum())
            fsat = float(Nsats) / cat.csize
            cat.attrs['fsat'] = fsat

    # mass distribution stats
    mass = cat[cat.attrs['halo_mass_key']].compute()
    logmass = numpy.log10(mass)
    avg_logmass = cat.comm.allreduce(logmass.sum()) / cat.csize
    sq_logmass = cat.comm.allreduce(((logmass - avg_logmass)**2).sum()) / cat.csize
    std_logmass = sq_logmass**0.5

    if cat.comm.rank == 0:
        if fsat is not None:
            cat.logger.info("satellite fraction: %.2f" % fsat)
        cat.logger.info("populated %d objects into %d halos" % (cat.csize, Nhalos))
        cat.logger.info("mean log10 halo mass: %.2f" % avg_logmass)
        cat.logger.info("std log10 halo mass: %.2f" % std_logmass)

def _enum_gal_types(galtab, types):
    """
    Enumerate the galaxy types as integers instead of strings.
    """
    if 'gal_type' in galtab.colnames:
        gal_type = numpy.zeros(len(galtab), dtype='i4')
        for i, gtype in enumerate(sorted(types[1:])):
            idx = galtab['gal_type'] == gtype
            gal_type[idx] = i+1
        galtab.replace_column('gal_type', gal_type)

def _test_for_objects(data):
    """
    Raise an exception if any of the columns have 'O' dtype.

    This is necessary because the 'O' dtype has an undefined size and cannot
    be scattered/gathered via MPI.
    """
    for col in data.colnames:
        if data.dtype[col] == 'O':
            raise TypeError("column '%s' is of type 'O'; must convert to integer or string" %col)
    return data



from runtests.mpi import MPITest
from nbodykit.io.fits import FITSFile

import os
import numpy
import tempfile
import pickle
import contextlib
import pytest

try: import fitsio
except ImportError: fitsio is None

@contextlib.contextmanager
def temporary_data(data='table'):
    """
    Write some temporary FITS data to disk
    """
    try:
        # generate data
        if data == 'table':
            dset = numpy.empty(1024, dtype=[('Position', ('f8', 3)), ('Mass', 'f8')])
            dset['Position'] = numpy.random.random(size=(1024, 3))
            dset['Mass'] = numpy.random.random(size=1024)
        else:
            dset = numpy.random.random(size=(1024, 3))

        # write to file
        tmpdir = tempfile.gettempdir()
        tmpfile = os.path.join(tmpdir, 'nbkit_tmp_data.fits')
        fitsio.write(tmpfile, dset, extname='Catalog')

        yield (dset, tmpfile)
    except:
        raise
    finally:
        os.unlink(tmpfile)


@MPITest([1])
@pytest.mark.skipif(fitsio is None, "fitsio is not installed")
def test_data(comm):

    with temporary_data() as (data, tmpfile):

        # read
        f = FITSFile(tmpfile)

        # check columns
        cols = ['Mass', 'Position']
        assert(all(col in cols for col in f.columns))

        # make sure data is the same
        for col in cols:
            numpy.testing.assert_almost_equal(data[col], f[col][:])

        # try a slice
        data2 = f.read(cols, 0, 512, 1)
        for col in cols:
            numpy.testing.assert_almost_equal(data[col][:512], data2[col])

        # check size
        numpy.testing.assert_equal(f.size, 1024)


@MPITest([1])
@pytest.mark.skipif(fitsio is None, "fitsio is not installed")
def test_string_ext(comm):

    with temporary_data() as (data, tmpfile):

        # read
        f = FITSFile(tmpfile, ext='Catalog')
        assert(all(col in ['Mass', 'Position'] for col in f.columns))
        assert(f.size == 1024)

        # read
        f = FITSFile(tmpfile, ext=1)
        assert(all(col in ['Mass', 'Position'] for col in f.columns))
        assert(f.size == 1024)

        # wrong ext
        with pytest.raises(ValueError):
            f = FITSFile(tmpfile, ext='WrongName')

@MPITest([1])
@pytest.mark.skipif(fitsio is None, "fitsio is not installed")
def test_wrong_ext(comm):

    with temporary_data() as (data, tmpfile):

        # no binary table data at 0
        with pytest.raises(ValueError):
            f = FITSFile(tmpfile, ext=0)

        # invalid ext number
        with pytest.raises(ValueError):
            f = FITSFile(tmpfile, ext=2)

@MPITest([1])
@pytest.mark.skipif(fitsio is None, "fitsio is not installed")
def test_no_tabular_data(comm):

    with temporary_data(data='image') as (data, tmpfile):

        # no binary table data
        with pytest.raises(ValueError):
            f = FITSFile(tmpfile)

        with pytest.raises(ValueError):
            f = FITSFile(tmpfile, ext=1)



from __future__ import absolute_import
# the future import is important. or in python 2.7 we try to
# import this module itself. Due to the unfortnate name conflict!

import numpy

from .base import FileType
from six import string_types
import json
from nbodykit.utils import JSONDecoder
from fnmatch import fnmatch


class Automatic: pass

class BigFile(FileType):
    """
    A file object to handle the reading of columns of data from
    a :mod:`bigfile` file.

    :mod:`bigfile` is a reproducible, massively parallel IO library for
    large, hierarchical datasets, and it is the default format of the
    `FastPM <https://github.com/rainwoodman/fastpm>`_ and the
    `MP-Gadget <https://github.com/bluetides-project/MP-Gadget>`_
    simulations.

    See also: https://github.com/rainwoodman/bigfile

    Parameters
    ----------
    path : str
        the name of the directory holding the bigfile data
    exclude : list of str, optional
        the data sets to exlude from loading within bigfile; default
        is the header. If any list is given, the name of the header column
        must be given too if it is not part of the data set. The names
        are shell glob patterns.

    header : str, or list, optional
        the path to the header; default is to use a column 'Header'.
        It is relative to the file, not the dataset.
        If a list is provided, the attributes is updated from the first entry to the last.

    dataset : str
        finding columns from a specific dataset in the bigfile;
        the default is start looking for columns from the root.
    """
    def __init__(self, path, exclude=None, header=Automatic, dataset='./'):

        if not dataset.endswith('/'): dataset = dataset + '/'

        import bigfile

        self.dataset = dataset
        self.path = path

        # store the attributes
        self.attrs = {}

        # the file path
        with bigfile.File(filename=path) as ff:
            columns = [block for block in ff[self.dataset].blocks]
            headers = self._find_headers(header, dataset, ff)

            if exclude is None:
                # by default exclude header only.
                exclude = headers

            if not isinstance(exclude, (list, tuple)):
                exclude = [exclude]

            columns = [
                column
                for column in set(columns) if not any(fnmatch(column, e) for e in exclude)
                ]

            ds = bigfile.Dataset(ff[self.dataset], columns)

            # set the data type and size
            self.dtype = ds.dtype
            self.size = ds.size

            headers = [ff[header] for header in headers]
            all_attrs = [ header.attrs for header in headers ]
            for attrs in all_attrs:
                # copy over the attrs
                for k in attrs.keys():

                    # load a JSON representation if str starts with json:://
                    if isinstance(attrs[k], string_types) and attrs[k].startswith('json://'):
                        self.attrs[k] = json.loads(attrs[k][7:], cls=JSONDecoder)
                    # copy over an array
                    else:
                        self.attrs[k] = numpy.array(attrs[k], copy=True)

    def _find_headers(self, header, dataset, ff):
        """ Find header from the file block by default. """
        if header is Automatic:
            header = ['Header', 'header', '.']

        if not isinstance(header, (tuple, list)):
            header = [header]

        r = []
        for h in header:
            if h in ff.blocks:
                if h not in r:
                    r.append(h)

        # append the dataset itself
        r.append(dataset.strip('/') + '/.')

        # shall not make the assertion here because header can be nested deep.
        # then not shown in ff.blocks. try catch may work better.
        #if not header in ff.blocks:
        #    raise KeyError("header block `%s` is not defined in the bigfile. Candidates can be `%s`"
        #            % (header, str(ff.blocks))

        return r

    def read(self, columns, start, stop, step=1):
        """
        Read the specified column(s) over the given range,
        as a dictionary

        'start' and 'stop' should be between 0 and :attr:`size`,
        which is the total size of the binary file (in particles)
        """
        import bigfile
        if isinstance(columns, string_types): columns = [columns]

        with bigfile.File(filename=self.path)[self.dataset] as f:
            ds = bigfile.Dataset(f, columns)
            return ds[start:stop][::step]



"""
The nbodykit lab, containing all of the necessary ingredients to use nbodykit.
"""
from mpi4py import MPI
import numpy

# sources
from nbodykit.source.catalog import *
from nbodykit.source.mesh import *

# algorithms
from nbodykit.algorithms import *

from nbodykit.batch import TaskManager
from nbodykit import cosmology
from nbodykit import CurrentMPIComm, GlobalCache
from nbodykit import transform
from nbodykit import filters
from nbodykit import io as IO

# HOD models
from nbodykit.hod import *

# the tutorials module
from nbodykit import tutorials



from runtests.mpi import MPITest
from nbodykit import set_options, GlobalCache, use_distributed, use_mpi
from numpy.testing import assert_array_equal
from nbodykit.lab import UniformCatalog

def setup():
    from distributed import LocalCluster, Client
    cluster = LocalCluster(n_workers=1, threads_per_worker=1, processes=False)
    use_distributed(Client(cluster))

def teardown():
    from mpi4py import MPI
    use_mpi(MPI.COMM_WORLD)

@MPITest([1])
def test_pickle(comm):
    import pickle
    cat = UniformCatalog(1e-3, 512, comm=comm)
    ss = pickle.dumps(cat)
    cat2 = pickle.loads(ss)
    
    assert_array_equal(cat['Position'], cat2['Position'])

@MPITest([1])
def test_save(comm):
    cat = UniformCatalog(1e-3, 512, comm=comm)

    import tempfile
    import shutil

    tmpfile = tempfile.mkdtemp()
    cat.save(tmpfile, dataset='data')
    shutil.rmtree(tmpfile)



from runtests.mpi import MPITest
from nbodykit.io.csv import CSVFile
import os
import numpy
import tempfile
import pickle
import pytest


@MPITest([1])
def test_no_trailing_newline(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data with blank lines
        data = numpy.array([[1, 1, 1, 1], [2, 2, 2, 2]], dtype='i4')
        ff.write(("1 1 1 1\n2 2 2 2").encode())
        ff.seek(0)

        # this will fail -- header should not be in file
        names =['a', 'b', 'c', 'd']
        f = CSVFile(path=ff.name, names=names, blocksize=1000)

        assert(f.size == 2)
        data2 = f.asarray()
        numpy.testing.assert_almost_equal(data, data2[:], decimal=7)

@MPITest([1])
def test_data(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)

        # read nrows
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVFile(path=ff.name, names=names, blocksize=100)

        # make sure data is the same
        data2 = f.asarray()
        numpy.testing.assert_almost_equal(data, data2[:], decimal=7)

        # make sure all the columns are there
        assert all(col in f for col in names)

@MPITest([1])
def test_pickle(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)

        # read nrows
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVFile(path=ff.name, names=names, blocksize=100)

        s = pickle.dumps(f)
        f2 = pickle.loads(s)

        numpy.testing.assert_almost_equal(f['a'][:], f2['a'][:])

@MPITest([1])
def test_slicing(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)

        # read nrows
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVFile(path=ff.name, names=names, blocksize=1000)

        # make sure data is the same (check only the first column her)
        for sl in [slice(0,10), slice(-10, -5), slice(0, 50, 2), slice(-50, None, 3)]:
            numpy.testing.assert_almost_equal(data[sl][:,0], f[sl]['a'], decimal=7)

@MPITest([1])
def test_comma_sep(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data, fmt='%.7e', delimiter=','); ff.seek(0)

        # this must fail
        names =['a', 'b', 'c', 'd', 'e']
        with pytest.raises(ValueError):
            f = CSVFile(path=ff.name, names=names, blocksize=100, delim_whitespace=True)

        # use , as delimiter
        f = CSVFile(path=ff.name, names=names, blocksize=100, delim_whitespace=False, sep=',')
        data2 = f.asarray()
        numpy.testing.assert_almost_equal(data, data2[:], decimal=7)


@MPITest([1])
def test_header_fail(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data with blank lines
        data = numpy.random.random(size=(100,5))
        ff.write(("# a b c d e\n").encode())
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)

        # this will fail -- header should not be in file
        names =['a', 'b', 'c', 'd', 'e']
        with pytest.raises(ValueError):
            f = CSVFile(path=ff.name, names=names, blocksize=1000)


@MPITest([1])
def test_blank_lines(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data with blank lines
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)

        # read
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVFile(path=ff.name, names=names, blocksize=100)

        # the right size
        assert f.size == len(f[:]), "mismatch between 'size' and data read from file"
        assert f.size == 100, "error reading with blank lines'"

@MPITest([1])
def test_dtype(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        ff.write(("\n\n\n").encode())
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)

        # specify the dtype as dict
        names =['a', 'b', 'c', 'd', 'e']
        dtype = {'a':'f4', 'b':'i8', 'c':'f16'}
        f = CSVFile(path=ff.name, names=names, blocksize=100, dtype=dtype)

        # make sure data is the same
        assert f.dtype['a'] == 'f4'
        assert f.dtype['b'] == 'i8'
        assert f.dtype['c'] == 'f16'

        # specify the dtype as dict
        f = CSVFile(path=ff.name, names=names, blocksize=100, dtype="f4")

        # make sure data is the same
        assert all(f.dtype[col] == 'f4' for col in names)


@MPITest([1])
def test_comments(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        ff.write(("# comment line 1\n# comment line 2\n").encode())
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)

        # this should raise an error
        names =['a', 'b', 'c', 'd', 'e']
        with pytest.raises(ValueError):
            f = CSVFile(path=ff.name, names=names, blocksize=1000)


@MPITest([1])
def test_skiprows(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)

        # read nrows
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVFile(path=ff.name, names=names, blocksize=1000, nrows=50, skiprows=25)

        # the right size
        assert f.size == len(f[:]), "mismatch between 'size' and data read from file"
        assert f.size == 50, "error combining 'skiprows' and 'nrows'"

        # make sure right portion of data was read
        data2 = f.asarray()
        numpy.testing.assert_almost_equal(data[25:75], data2[:], decimal=7)

@MPITest([1])
def test_nrows(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data); ff.seek(0)

        # read nrows
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVFile(path=ff.name, names=names, blocksize=1000, nrows=50)

        # the right size
        assert f.size == len(f[:]), "mismatch between 'size' and data read from file"
        assert f.size == 50, "error reading 'nrows'"

        # make sure right portion of data was read
        data2 = f.asarray()
        numpy.testing.assert_almost_equal(data[:50], data2[:], decimal=7)

@MPITest([1])
def test_usecols(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data); ff.seek(0)

        # read usecols
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVFile(path=ff.name, names=names, blocksize=1000, usecols=['a', 'c', 'e'])

        assert f.columns == ['a', 'c', 'e'], "error using 'usecols'"

@MPITest([1])
def test_wrong_names(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data); ff.seek(0)

        # wrong number of columns
        names =['a', 'b', 'c']
        with pytest.raises(ValueError):
            f = CSVFile(path=ff.name, names=names, blocksize=1000)


@MPITest([1])
def test_invalid_keywords(comm):

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data); ff.seek(0)

        # the bad keywords
        bad_kws = {'index_col':True, 'header':True, 'skiprows':[0,1,2], 'skipfooter':True, 'comment':'#'}

        # try each bad kewyord
        names =['a', 'b', 'c', 'd', 'e']
        for k,v in bad_kws.items():
            with pytest.raises(ValueError):
                f = CSVFile(path=ff.name, names=names, blocksize=1000, **{k:v})



from six import string_types
import numpy
import logging
from abc import abstractmethod
from nbodykit import _global_options

class FileType(object):
    """
    An abstract base class representing a file object.

    Users should subclass this class and implement the :func:`read`
    function, responsible for reading data from the specific file type.
    """
    logger = logging.getLogger("FileType")

    @abstractmethod
    def read(self, columns, start, stop, step=1):
        """
        Read the specified column(s) over the given range,
        returning a structured numpy array

        Parameters
        ----------
        columns : str, list of str
            the name of the column(s) to return
        start : int
            the row integer to start reading at
        stop : int
            the row integer to stop reading at
        step : int, optional
            the step size to use when reading; default is 1

        Returns
        -------
        data : array_like
            a numpy structured array holding the requested data
        """
        pass

    @property
    def columns(self):
        """
        A list of the names of the columns in the file.

        This defaults to the named fields in the file's :attr:`dtype`
        attribute, but differ from this if a view of the file has been
        returned with :func:`asarray`
        """
        try:
            return self._columns
        except AttributeError:
            return list(self.dtype.names)

    @columns.setter
    def columns(self, val):
        self._columns = val

    @property
    def ncol(self):
        """
        The number of data columns in the file.
        """
        return len(self.columns)

    @property
    def shape(self):
        """
        The shape of the file, which defaults to ``(size, )``

        Multiple dimensions can be introduced into the shape if
        a view of the file has been returned with :func:`asarray`
        """
        try:
            return self._shape
        except AttributeError:
            return (self.size,)

    @shape.setter
    def shape(self, val):
        self._shape = val

    @property
    def size(self):
        """
        The size of the file, i.e., number of rows
        """
        try:
            return self._size
        except:
            name = self.__class__.__name__
            raise AttributeError("please set the ``size`` attribute when initializing the '%s' class" %name)

    @size.setter
    def size(self, val):
        self._size = val

    @property
    def dtype(self):
        """
        A :class:`numpy.dtype` object holding the data types of each column in the file.
        """
        try:
            return self._dtype
        except:
            name = self.__class__.__name__
            raise AttributeError("please set the ``dtype`` attribute when initializing the '%s' class" %name)

    @dtype.setter
    def dtype(self, val):
        self._dtype = val

    def __len__(self):
        return self.size

    def __iter__(self):
        return iter(self.keys())

    def __repr__(self):
        args = (self.__class__.__name__, self.path, self.dataset if hasattr(self, 'dataset') else "None", self.ncol, self.shape)
        return "%s(path=%s, dataset=%s, ncolumns=%d, shape=%s>" % args

    def __contains__(self, col):
        return col in self.columns

    def keys(self):
        """
        Aliased function to return :attr:`columns`
        """
        return list(self.columns)

    def __getitem__(self, s):
        """
        This function provides numpy-like array indexing of the file object.

        It supports:

        #. integer, slice-indexing similar to arrays
        #. string indexing using column names in :func:`keys`
        #. array-like indexing using integer lists or boolean arrays

        .. note::

            If a single column is being returned, a numpy array
            holding the data is returned, rather than a structured
            array with only a single field.
        """
        # dont call asarray unless we have a single string index
        asarray = False
        if isinstance(s, string_types):
            s = [s]
            asarray = True

        # if index is a list, it should contain a series of column names
        # this will return a "view" of the file, slicing the data type
        # to include only the requested columns
        if isinstance(s, list) and all(isinstance(k, string_types) for k in s):

            # empty slice
            if not len(s):
                raise IndexError("no columns selected in slice")

            # crash if the dtype has no fields
            if self.dtype.names is None:
                raise IndexError(("cannot access view of specific columns after `asarray()` "
                                  "has been called; use integer array indexing instead"))

            # all strings must be valid column names
            if not all(ss in self.keys() for ss in s):
                invalid = [col for col in s if s not in self.keys()]
                raise IndexError("invalid string keys: %s; run keys() for valid options" %str(invalid))

            # create a new object, with slice of dtype
            # FIXME: the new object shall be a different type.
            obj = object.__new__(self.__class__)
            obj.dtype = numpy.dtype([(col, self.dtype[col]) for col in s])
            obj.size = self.size

            # set the owner of the underlying memory
            if getattr(self, 'base', None) is not None:
                obj.base = self.base
            else:
                obj.base = self

            # return the single numpy array if only a
            # single column was asked for
            if len(s) == 1 and asarray: obj = obj.asarray()

            return obj

        # tuple for indices in multiple dimensions
        # this can either be of length 1 or 2
        second_axis_index = None
        if isinstance(s, tuple):

            # verify the tuple shape
            if len(s) > len(self.shape):
                args = len(self.shape), len(s)
                raise IndexError("file dimension is %d, but you supplied tuple of length %d" %args)

            if len(s) == 1:
                s = s[0]
            elif len(s) == 2:
                s, second_axis_index = s
            else:
                raise IndexError("tuple index '%s' not understood" %str(s))

        # call the read function over the desired row range
        # if we don't own memory, return from 'base' attribute
        if getattr(self, 'base', None) is None:
            memown = self
        else:
            memown = self.base

        # a list here means we are dealing with array-like indexing
        if isinstance(s, list):
            s = numpy.array(s)

        # do array-like indexing
        if isinstance(s, numpy.ndarray):

            # make all integers indexing positive
            if s.dtype == numpy.integer:
                s[s < 0] += len(self)

            # read the full desired slice in consecutive chunks
            toret = numpy.concatenate([memown.read(self.keys(),*sl) for sl in find_slice_chunks(s)])

        # slice contiguous chunk via (start, stop, step)
        else:
            # input is integer
            if isinstance(s, int):
                if s < 0: s += self.size
                start, stop, step = s, s+1, 1
            # input is a slice
            elif isinstance(s, slice):
                start, stop, step = s.indices(self.size)
            else:
                raise IndexError("index '%s' not understood - should be an integer or slice" %str(s))

            # call the read function over the desired row range
            toret = memown.read(self.keys(), start, stop, step)

        # if file has no named fields, then
        # try to view the output as a single numpy array
        if len(self.dtype) == 0:
            try:
                # FIXME: shall we ensure read always returns C-contiguous?
                toret = numpy.ascontiguousarray(toret).view(self.dtype)
                if len(self.shape) > 1:
                    toret = toret.reshape((-1, self.shape[1]))
            except Exception as e:
                raise ValueError("error trying to view slice as a single numpy array: %s" %str(e))

        # if we have an index for the second dimension
        # then slice the return value
        if second_axis_index is not None:
            toret = toret[:,second_axis_index]

        return toret

    def asarray(self):
        """
        Return a view of the file, where the fields of the
        structured array are stacked in columns of a single
        numpy array

        Examples
        --------

        Start with a file object with three named columns,
        ``ra``, ``dec``, and ``z``

        >>> ff.dtype
        dtype([('ra', '<f4'), ('dec', '<f4'), ('z', '<f4')])
        >>> ff.shape
        (1000,)
        >>> ff.columns
        ['ra', 'dec', 'z']
        >>> ff[:3]
        array([(235.63442993164062, 59.39099884033203, 0.6225500106811523),
               (140.36181640625, -1.162310004234314, 0.5026500225067139),
               (129.96627807617188, 45.970130920410156, 0.4990200102329254)],
              dtype=(numpy.record, [('ra', '<f4'), ('dec', '<f4'), ('z', '<f4')]))

        Select a subset of columns and switch the ordering
        and convert output to a single numpy array

        >>> x = ff[['dec', 'ra']].asarray()
        >>> x.dtype
        dtype('float32')
        >>> x.shape
        (1000, 2)
        >>> x.columns
        ['dec', 'ra']
        >>> x[:3]
        array([[  59.39099884,  235.63442993],
               [  -1.16231   ,  140.36181641],
               [  45.97013092,  129.96627808]], dtype=float32)

        Now, select only the first column (``dec``)

        >>> dec = x[:,0]
        >>> dec[:3]
        array([ 59.39099884,  -1.16231   ,  45.97013092], dtype=float32)

        Returns
        -------
        FileType :
            a file object that will return a numpy array with
            the columns representing the fields
        """
        # no named fields --> crash
        if not len(self.dtype):
            raise ValueError("no named dtype fields to convert to numpy array")

        # multiple vector dtypes --> crash
        if len(self.dtype) > 1 and any(len(self.dtype[col].shape) for col in self.dtype.names):
            raise ValueError("cannot convert multiple vector data types to numpy array")

        # different dtypes --> crash
        if any(self.dtype[col].base != self.dtype[0].base for col in self.dtype.names):
            raise ValueError("cannot convert data types of different types to single numpy array")

        # create the new object
        # FIXME: the new object shall be a different type.
        obj = object.__new__(self.__class__)

        # the second axis of the shape
        if len(self.dtype) == 1:
            subshape = self.dtype[0].shape
        else:
            subshape = (len(self.dtype),)

        obj.dtype   = self.dtype[0].base
        obj.columns = list(self.columns)
        obj.shape   = (self.size, ) + subshape
        obj.size    = self.size
        if getattr(self, 'base', None) is not None:
            obj.base = self.base
        else:
            obj.base = self

        return obj

    def get_dask(self, column, blocksize=None):
        """
        Return the specified column as a dask array, which
        delays the explicit reading of the data until
        :meth:`dask.compute` is called

        The dask array is chunked into blocks of size `blocksize`

        Parameters
        ----------
        column : str
            the name of the column to return
        blocksize : int, optional
            the size of the chunks in the dask array

        Returns
        -------
        :class:`dask.array.Array` :
            the dask array holding the column, which computes the
            necessary functions to read the data, but delays evaluating
            until the user specifies
        """
        if blocksize is None:
            blocksize = _global_options['dask_chunk_size']

        if column not in self:
            raise ValueError("'%s' is not a valid column; run keys() for valid options" %column)

        import dask.array as da
        return da.from_array(self[column], chunks=blocksize)


def find_slice_chunks(index):
    """
    A generator to yield (start, stop, step) tuples
    which will correspond to the input selection index

    ``index`` can be either a boolen index, or a list of integers
    specifying the rows to include

    Parameters
    ----------
    index : array_like
        either a boolean array, indicating which rows to select,
        or integers specifying which rows to include

    Yields
    ------
    (start, stop, step) : tuple of int
        the slice integers to read, corresponding to a valid spart of the
        selection index
    """
    from itertools import groupby
    from operator import itemgetter

    if isinstance(index, list):
        index = numpy.array(index)

    # handle boolean index
    if index.dtype == '?':
        vals, N = zip(*[(k, sum(1 for i in g)) for k,g in groupby(index)])
        N = numpy.cumsum(N)

        for i, v in enumerate(vals):
            if v:
                ilow = i-1
                if ilow < 0: Nlow = 0
                else: Nlow = N[ilow]

                yield (Nlow, N[i], 1)

    # handle integer index
    else:
        N = []
        for k,g in groupby(enumerate(index), lambda x: x[0]-x[1]):
            N.append(list(map(itemgetter(1), g)))

        for xx in N:
            yield (xx[0], xx[-1]+1, 1)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_array_equal, assert_allclose
import kdcount.correlate as correlate
import os
import pytest

setup_logging()

def gather_data(source, name):
    return numpy.concatenate(source.comm.allgather(source[name].compute()), axis=0)

def generate_sim_data(seed, dtype, comm):
    return UniformCatalog(nbar=3e-6, BoxSize=512., seed=seed, dtype=dtype, comm=comm)

def generate_survey_data(seed, dtype, comm):
    cosmo = cosmology.Planck15
    s = RandomCatalog(1000, seed=seed, comm=comm)

    s['Redshift'] = s.rng.normal(loc=0.5, scale=0.1).astype(dtype)
    s['RA'] = s.rng.uniform(low=0, high=360).astype(dtype)
    s['DEC'] = s.rng.uniform(low=-60, high=60.).astype(dtype)
    s['Position'] = transform.SkyToCartesian(s['RA'], s['DEC'], s['Redshift'], cosmo=cosmo).astype(dtype)

    return s

def reference_paircount(pos1, w1, redges, boxsize, pos2=None, w2=None, los=2):
    """Reference pair counting via kdcount"""
    # make the trees
    tree1 = correlate.points(pos1, boxsize=boxsize, weights=w1)
    if pos2 is None:
        tree2 = tree1
    else:
        tree2 = correlate.points(pos2, boxsize=boxsize, weights=w2)

    # do the paircount
    bins = correlate.RBinning(redges)
    pc = correlate.paircount(tree1, tree2, bins, np=0, compute_mean_coords=True)
    return numpy.nan_to_num(pc.pair_counts), numpy.nan_to_num(pc.mean_centers), pc.sum1

@MPITest([1, 3])
def test_sim_periodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, dtype='f8', comm=comm)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    r = SimulationBoxPairCount('1d', source, redges, periodic=True, weight='Weight')

    # cannot compute r=0
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, numpy.linspace(0, 10.0, 10))

    pos = gather_data(source, "Position")
    w = gather_data(source, "Weight")

    # verify with kdcount
    npairs, ravg, wsum = reference_paircount(pos, w, redges, source.attrs['BoxSize'])
    assert_allclose(ravg, r.pairs['r'])
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])

    # test save
    r.save('paircount-test.json')
    r2 = SimulationBoxPairCount.load('paircount-test.json', comm=comm)
    assert_array_equal(r.pairs.data, r2.pairs.data)

    if comm.rank == 0: os.remove('paircount-test.json')

@MPITest([1, 3])
def test_sim_nonperiodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, dtype='f8', comm=comm)
    source['Weight'] = numpy.random.random(size=len(source))

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    r = SimulationBoxPairCount('1d', source, redges, periodic=False, weight='Weight')

    pos = gather_data(source, "Position")
    w = gather_data(source, "Weight")

    # verify with kdcount
    npairs, ravg, wsum = reference_paircount(pos, w, redges, None)
    assert_allclose(ravg, r.pairs['r'])
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])


@MPITest([1, 3])
def test_sim_periodic_cross(comm):

    # generate data
    first = generate_sim_data(seed=42, dtype='f4', comm=comm)
    second = generate_sim_data(seed=84, dtype='f8', comm=comm)

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the paircount
    r = SimulationBoxPairCount('1d', first, redges, second=second, periodic=True)

    pos1 = gather_data(first, "Position")
    pos2 = gather_data(second, "Position")

    # verify with kdcount
    npairs, ravg, wsum = reference_paircount(pos1, None, redges, first.attrs['BoxSize'], pos2=pos2)
    assert_allclose(ravg, r.pairs['r'], rtol=1e-6)
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])

@MPITest([1])
def test_bad_los1(comm):
    source = generate_sim_data(seed=42, dtype='f8', comm=comm)
    redges = numpy.linspace(10, 150, 10)

    # should be 'x', 'y', 'z'
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, redges, los='a')

@MPITest([1])
def test_bad_los2(comm):
    source = generate_sim_data(seed=42, dtype='f8', comm=comm)
    redges = numpy.linspace(10, 150, 10)

    # should be [0,1,2]
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, redges, los=3)

@MPITest([1])
def test_bad_los3(comm):
    source = generate_sim_data(seed=42, dtype='f8', comm=comm)
    redges = numpy.linspace(10, 150, 10)

    return # stop early to see if illegal instruction is gone.
    # negative okay
    r = SimulationBoxPairCount('1d', source, redges, los=-1)

@MPITest([1])
def test_bad_los4(comm):
    source = generate_sim_data(seed=42, dtype='f8', comm=comm)
    redges = numpy.linspace(10, 150, 10)

    # vector is bad
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, redges, los=[0,0,1])

@MPITest([1])
def test_bad_rmax(comm):

    source = generate_sim_data(seed=42, dtype='f8', comm=comm)
    redges = numpy.linspace(10, 400., 10)

    # rmax is too big
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, redges, periodic=True)

@MPITest([1])
def test_noncubic_box(comm):

    source = UniformCatalog(nbar=3e-6, BoxSize=[512., 512., 256], seed=42, comm=comm)
    redges = numpy.linspace(10, 100., 10)

    # cannot do non-cubic boxes
    with pytest.raises(NotImplementedError):
        r = SimulationBoxPairCount('1d', source, redges, periodic=True)

@MPITest([1])
def test_bad_boxsize(comm):

    # uniform source of particles
    first = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42, comm=comm)
    second = UniformCatalog(nbar=3e-5, BoxSize=256., seed=42, comm=comm)

    # make the bin edges
    redges = numpy.linspace(10, 50., 10)

    # box sizes are different
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', first, redges, second=second)

    # specified different value
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', second, redges, BoxSize=300.)

    # no boxSize
    del first.attrs['BoxSize']
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', first, redges)

@MPITest([1, 4])
def test_survey_auto(comm):

    cosmo = cosmology.Planck15

    # random particles
    source = generate_survey_data(seed=42, dtype='f8', comm=comm)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    r = SurveyDataPairCount('1d', source, redges, cosmo, weight='Weight')

    # cannot compute r=0
    with pytest.raises(ValueError):
        r = SurveyDataPairCount('1d', source, numpy.linspace(0, 10.0, 10), cosmo)

    pos = gather_data(source, 'Position')
    w = gather_data(source, 'Weight')

    # verify with kdcount
    npairs, ravg, wsum = reference_paircount(pos, w, redges, None)
    assert_allclose(ravg, r.pairs['r'])
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])

@MPITest([1, 4])
def test_survey_auto_endianess(comm):

    cosmo = cosmology.Planck15

    # random particles
    source = generate_survey_data(seed=42, dtype='f8', comm=comm)

    source['RA'] = source['RA'].astype('>f8')
    source['DEC'] = source['DEC'].astype('>f8')
    source['Redshift'] = source['Redshift'].astype('>f8')
    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform().astype('>f8')
    source['Position'] = source['Position'].astype('>f8')

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    r = SurveyDataPairCount('1d', source, redges, cosmo, weight='Weight')

    # cannot compute r=0
    with pytest.raises(ValueError):
        r = SurveyDataPairCount('1d', source, numpy.linspace(0, 10.0, 10), cosmo)

    pos = gather_data(source, 'Position')
    w = gather_data(source, 'Weight')

    # verify with kdcount
    npairs, ravg, wsum = reference_paircount(pos, w, redges, None)
    assert_allclose(ravg, r.pairs['r'])
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])

@MPITest([1, 4])
def test_survey_cross(comm):

    cosmo = cosmology.Planck15

    # random particles
    first = generate_survey_data(seed=42, dtype='f4', comm=comm)
    first['Weight'] = first.rng.uniform(dtype='f4')
    # mismatched dtype shouldn't fail
    second = generate_survey_data(seed=84, dtype='f8', comm=comm)
    second['Weight'] = second.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the paircount
    r = SurveyDataPairCount('1d', first, redges, cosmo, second=second)

    pos1 = gather_data(first, 'Position')
    pos2 = gather_data(second, 'Position')
    w1 = gather_data(first, 'Weight')
    w2 = gather_data(second, 'Weight')

    # verify with kdcount
    npairs, ravg, wsum = reference_paircount(pos1, w1, redges, None, pos2=pos2, w2=w2)
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])
    # the error can be larger due to single precision positions in one of
    # the dataset
    assert_allclose(ravg, r.pairs['r'], rtol=1e-5)

@MPITest([1])
def test_survey_missing_columns(comm):
    source = generate_survey_data(seed=42, dtype='f8', comm=comm)
    redges = numpy.linspace(10, 150, 10)

    # missing column
    with pytest.raises(ValueError):
        r = SurveyDataPairCount('1d', source, redges, cosmology.Planck15, ra='BAD')

@MPITest([1])
def test_bad_mode(comm):
    source = generate_survey_data(seed=42, dtype='f8', comm=comm)
    redges = numpy.linspace(10, 150, 10)

    # missing column
    with pytest.raises(ValueError):
        r = SurveyDataPairCount('bad mode', source, redges, cosmology.Planck15)

@MPITest([1, 4])
def test_corrfunc_exception(comm):

    pos = numpy.zeros((100,3))
    cat = ArrayCatalog({'Position':pos}, comm=comm)

    redges = numpy.linspace(0.01, 0.1, 10)

    # corrfunc will throw an error due to bad input data
    with pytest.raises(Exception):
        r = SimulationBoxPairCount('1d', cat, redges, periodic=False, BoxSize=[1., 1., 1,])

@MPITest([1, 4])
def test_missing_corrfunc(comm):

    from nbodykit.algorithms.pair_counters.corrfunc.base import MissingCorrfuncError

    with pytest.raises(Exception):
        raise MissingCorrfuncError()

@MPITest([1])
def test_missing_Position(comm):

    pos = numpy.zeros((100,3))
    cat = ArrayCatalog({'MissingPosition':pos}, comm=comm)
    redges = numpy.linspace(0.01, 0.1, 10)

    # cat is missing Position
    with pytest.raises(Exception):
        r = SimulationBoxPairCount('1d', cat, redges, periodic=False, BoxSize=[1., 1., 1,])

    cat2 = cat.copy()
    cat['Position'] = cat['MissingPosition']

    # cat2 is missing Position
    with pytest.raises(Exception):
        r = SimulationBoxPairCount('1d', cat, redges, second=cat2, periodic=False, BoxSize=[1., 1., 1,])



from nbodykit.source.mesh import MultipleSpeciesCatalogMesh
from nbodykit.source.mesh import CatalogMesh
from nbodykit.utils import attrs_to_dict
import logging
import numpy

class FKPCatalogMesh(MultipleSpeciesCatalogMesh):
    """
    A subclass of
    :class:`~nbodykit.source.catalogmesh.species.MultipleSpeciesCatalogMesh`
    designed to paint a :class:`~nbodykit.source.catalog.fkp.FKPCatalog` to
    a mesh.

    The multiple species here are ``data`` and ``randoms`` CatalogSource
    objects, where ``randoms`` is a catalog of randomly distributed objects
    with no instrinsic clustering that defines the survey volume.

    The position of the catalogs are re-centered to the ``[-L/2, L/2]``
    where ``L`` is the size of the Cartesian box.

    Parameters
    ----------
    source : CatalogSource
        the input catalog that we wish to interpolate to a mesh
    BoxSize :
        the size of the box
    Nmesh : int, 3-vector
        the number of cells per mesh side
    dtype : str
        the data type of the values stored on mesh
    selection : str
        column in ``source`` that selects the subset of particles to grid
        to the mesh
    comp_weight : str
        the completeness weight column name
    fkp_weight : str
        the FKP weight column name
    nbar : str
        the n(z) column name
    position : str, optional
        column in ``source`` specifying the position coordinates; default
        is ``Position``
    """
    logger = logging.getLogger('FKPCatalogMesh')

    def __init__(self, source, BoxSize, BoxCenter, Nmesh, dtype, selection,
                    comp_weight, fkp_weight, nbar, value='Value',
                    position='Position', interlaced=False,
                    compensated=False, resampler='cic'):

        from .catalog import FKPCatalog
        if not isinstance(source, FKPCatalog):
            raise TypeError("the input source for FKPCatalogMesh must be a FKPCatalog")

        uncentered_position = position
        position = '_RecenteredPosition'
        weight = '_TotalWeight'

        self.attrs.update(source.attrs)

        self.recenter_box(BoxSize, BoxCenter)

        MultipleSpeciesCatalogMesh.__init__(self, source=source,
                        BoxSize=BoxSize, Nmesh=Nmesh,
                        dtype=dtype, weight=weight, value=value, selection=selection, position=position,
                        interlaced=interlaced, compensated=compensated, resampler=resampler)

        self._uncentered_position = uncentered_position
        self.comp_weight = comp_weight
        self.fkp_weight = fkp_weight
        self.nbar = nbar

    def __getitem__(self, key):
        """
        If indexed by a species name, return a CatalogMesh object holding
        only the data columns for that species with the same parameters as
        the current object.

        If not a species name, this has the same behavior as
        :func:`CatalogSource.__getitem__`.
        """
        assert key in self.source.species, "the species is not defined in the source"

        # CatalogSource holding only requested species
        cat = self.source[key]

        assert cat.comm is self.comm

        # view as a catalog mesh
        mesh = CatalogMesh(cat,
                BoxSize=self.attrs['BoxSize'],
                Nmesh=self.attrs['Nmesh'],
                dtype=self.dtype,
                Weight=self.TotalWeight(key),
                Value=cat[self.value],
                Selection=cat[self.selection],
                Position=self.RecenteredPosition(key),
                interlaced=self.interlaced,
                compensated=self.compensated,
                resampler=self.resampler,
            )

        return mesh

    def recenter_box(self, BoxSize, BoxCenter):
        """
        Re-center the box by applying the new box center to the column specified
        by :attr:`position`.

        This ensures that the position column is always re-centered to
        ``[-L/2,L/2]`` where ``L`` is the BoxSize.
        """
        # check input type
        BoxSize = numpy.ones(3) * BoxSize
        BoxCenter = numpy.ones(3) * BoxCenter

        # update meta-data
        for val, name in zip([BoxSize, BoxCenter], ['BoxSize', 'BoxCenter']):
            self.attrs[name] = val


    def to_real_field(self):
        r"""
        Paint the FKP density field, returning a ``RealField``.

        Given the ``data`` and ``randoms`` catalogs, this paints:

        .. math::

            F(x) = w_\mathrm{fkp}(x) * [w_\mathrm{comp}(x)*n_\mathrm{data}(x) -
                        \alpha * w_\mathrm{comp}(x)*n_\mathrm{randoms}(x)]


        This computes the following meta-data attributes in the process of
        painting, returned in the :attr:`attrs` attributes of the returned
        RealField object:

        - randoms.W, data.W :
            the weighted sum of randoms and data objects; see
            :func:`weighted_total`
        - alpha : float
            the ratio of ``data.W`` to ``randoms.W``
        - randoms.norm, data.norm : float
            the power spectrum normalization; see :func:`normalization`
        - randoms.shotnoise, data.shotnoise: float
            the shot noise for each sample; see :func:`shotnoise`
        - shotnoise : float
            the total shot noise, equal to the sum of ``randoms.shotnoise``
            and ``data.shotnoise``
        - randoms.num_per_cell, data.num_per_cell : float
            the mean number of weighted objects per cell for each sample
        - num_per_cell : float
            the mean number of weighted objects per cell

        For further details on the meta-data, see
        :ref:`the documentation <fkp-meta-data>`.

        Returns
        -------
        :class:`~pmesh.pm.RealField` :
            the field object holding the FKP density field in real space
        """

        attrs = {}

        # determine alpha, the weighted number ratio
        for name in self.source.species:
            attrs[name+'.W'] = self.weighted_total(name)

        attrs['alpha'] = attrs['data.W'] / attrs['randoms.W']

        # paint the data
        real = self['data'].to_real_field(normalize=False)
        real.attrs.update(attrs_to_dict(real, 'data.'))
        if self.comm.rank == 0:
            self.logger.info("data painted.")

        if self.source['randoms'].csize > 0:

            # paint the randoms
            real2 = self['randoms'].to_real_field(normalize=False)

            # normalize the randoms by alpha
            real2[:] *= -1. * attrs['alpha']

            if self.comm.rank == 0:
                self.logger.info("randoms painted.")

            real[:] += real2[:]
            real.attrs.update(attrs_to_dict(real2, 'randoms.'))

        # divide by volume per cell to go from number to number density
        vol_per_cell = (self.pm.BoxSize/self.pm.Nmesh).prod()
        real[:] /= vol_per_cell

        if self.comm.rank == 0:
            self.logger.info("volume per cell is %g" % vol_per_cell)

        # remove shot noise estimates (they are inaccurate in this case)
        real.attrs.update(attrs)
        real.attrs.pop('data.shotnoise', None)
        real.attrs.pop('randoms.shotnoise', None)

        return real

    def RecenteredPosition(self, name):
        """
        The Position of the objects, re-centered on the mesh to
        the range ``[-BoxSize/2, BoxSize/2]``.

        This subtracts ``BoxCenter`` from :attr:`attrs` from the original
        position array.
        """
        assert name in ['data', 'randoms']
        return self.source[name][self._uncentered_position] - self.attrs['BoxCenter']

    def TotalWeight(self, name):
        """
        The total weight for the mesh is the completenes weight times
        the FKP weight.
        """
        assert name in ['data', 'randoms']
        return self.source[name][self.comp_weight] * self.source[name][self.fkp_weight]

    def weighted_total(self, name):
        r"""
        Compute the weighted total number of objects, using either the
        ``data`` or ``randoms`` source:

        This is the sum of the completeness weights:

        .. math::

            W = \sum w_\mathrm{comp}
        """
        # the selection
        sel = self.source.compute(self.source[name][self.selection])

        # the selected mesh for "name"
        selected = self.source[name][sel]

        # sum up completeness weights
        wsum = self.source.compute(selected[self.comp_weight].sum())
        return self.comm.allreduce(wsum)



from nbodykit.lab import *
from nbodykit import setup_logging
from runtests.mpi import MPITest

import pytest

# debug logging
setup_logging("debug")

from nbodykit.filters import TopHat, Gaussian

@MPITest([1])
def test_tophat(comm):
    from nbodykit.base.mesh import MeshFilter

    cosmo = cosmology.Planck15

    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)
    source2 = source.apply(TopHat(8))

    r2 = source2.paint()
    # FIXME: add numerical assertions

@MPITest([1])
def test_tophat(comm):
    from nbodykit.base.mesh import MeshFilter

    cosmo = cosmology.Planck15

    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)
    source2 = source.apply(Gaussian(8))
    r2 = source2.paint()
    # FIXME: add numerical assertions



from nbodykit import GlobalCache
from nbodykit.lab import UniformCatalog

from runtests.mpi import MPITest

@MPITest([1])
def test_cache(comm):
    cat = UniformCatalog(nbar=10000, BoxSize=1.0, comm=comm)
    cat['test'] = cat['Position'] ** 5
    test = cat['test'].compute()

    # cache should no longer be empty
    cache = GlobalCache.get()

    assert cache.cache.total_bytes > 0

    # resize
    cache.cache.available_bytes = 100
    cache.cache.shrink()

    assert cache.cache.total_bytes < 100



from nbodykit.base.catalog import CatalogSource
from nbodykit.io.stack import FileStack
from nbodykit import CurrentMPIComm
from nbodykit import io
from nbodykit.extern import docrep

from six import string_types
import textwrap
import os

__all__ = ['FileCatalogFactory', 'FileCatalogBase',
           'CSVCatalog', 'BinaryCatalog', 'BigFileCatalog',
           'HDFCatalog', 'TPMBinaryCatalog', 'Gadget1Catalog', 'FITSCatalog']

class FileCatalogBase(CatalogSource):
    """
    Base class to create a source of particles from a
    single file, or multiple files, on disk.

    Files of a specific type should be subclasses of this class.

    Parameters
    ----------
    filetype : subclass of :class:`~nbodykit.io.base.FileType`
        the file-like class used to load the data from file; should be a
        subclass of :class:`nbodykit.io.base.FileType`
    args : tuple, optional
        the arguments to pass to the ``filetype`` class when constructing
        each file object
    kwargs : dict, optional
        the keyword arguments to pass to the ``filetype`` class when
        constructing each file object
    comm : MPI Communicator, optional
        the MPI communicator instance; default (``None``) sets to the
        current communicator
    """
    @CurrentMPIComm.enable
    def __init__(self, filetype, args=(), kwargs={}, comm=None):

        self.comm = comm
        self.filetype = filetype

        # bcast the FileStack
        if self.comm.rank == 0:
            self._source = FileStack(filetype, *args, **kwargs)
        else:
            self._source = None
        self._source = self.comm.bcast(self._source)

        # compute the size
        start = self.comm.rank * self._source.size // self.comm.size
        end = (self.comm.rank  + 1) * self._source.size // self.comm.size
        self._size = end - start

        # update the meta-data
        self.attrs.update(self._source.attrs)

        if self.comm.rank == 0:
            self.logger.info("Extra arguments to FileType: %s %s" % (str(args), str(kwargs)))

        CatalogSource.__init__(self, comm=comm)

    def __repr__(self):
        path = self._source.path
        name = self.__class__.__name__
        args = (name, self.size, repr(self._source))

        return "%s(size=%d, %s)" % args

    @property
    def hardcolumns(self):
        """
        The union of the columns in the file and any transformed columns.
        """
        defaults = CatalogSource.hardcolumns.fget(self)
        return list(self._source.dtype.names) + defaults

    def get_hardcolumn(self, col):
        """
        Return a column from the underlying file source.

        Columns are returned as dask arrays.
        """
        if col in self._source.dtype.names:
            start = self.comm.rank * self._source.size // self.comm.size
            end = (self.comm.rank  + 1) * self._source.size // self.comm.size
            return self._source.get_dask(col)[start:end]
        else:
            return CatalogSource.get_hardcolumn(self, col)


def _make_docstring(filetype, examples):
    """
    Internal function to generate the doc strings for the built-in
    CatalogSource objects that rely on :mod:`nbodykit.io` classes
    to read data from disk.
    """

    qualname = '%s.%s' %(filetype.__module__, filetype.__name__)
    __doc__ = """
A CatalogSource that uses :class:`~{qualname}` to read data from disk.

Multiple files can be read at once by supplying a list of file
names or a glob asterisk pattern as the ``path`` argument. See
:ref:`reading-multiple-files` for examples.

Parameters
----------
%(test.parameters)s
comm : MPI Communicator, optional
    the MPI communicator instance; default (``None``) sets to the
    current communicator
attrs : dict, optional
    dictionary of meta-data to store in :attr:`attrs`
""".format(qualname=qualname)

    if examples is not None:
        __doc__ += """
Examples
--------
Please see :ref:`the documentation <%s>` for examples.
""" %examples

    # get the Parameters from the IO libary class
    d = docrep.DocstringProcessor()
    d.get_sections(d.dedents(filetype.__doc__), 'test', ['Parameters'])
    return d.dedents(__doc__)

def FileCatalogFactory(name, filetype, examples=None):
    """
    Factory method to create a :class:`~nbodykit.base.catalog.CatalogSource`
    that uses a subclass of :mod:`nbodykit.io.base.FileType` to read
    data from disk.

    Parameters
    ----------
    name : str
        the name of the catalog class to create
    filetype : subclass of :class:`nbodykit.io.base.FileType`
        the subclass of the FileType that reads a specific type of data
    examples : str, optional
        if given, a documentation cross-reference link where examples can be
        found

    Returns
    -------
    subclass of :class:`FileCatalogBase` :
        the ``CatalogSource`` object that reads data using ``filetype``
    """
    def __init__(self, *args, **kwargs):
        comm = kwargs.pop('comm', None)
        attrs = kwargs.pop('attrs', {})
        FileCatalogBase.__init__(self, filetype=filetype, args=args, kwargs=kwargs, comm=comm)
        self.attrs.update(attrs)

    # make the doc string for this class
    __doc__ = _make_docstring(filetype, examples)

    # make the new class object and return it
    newclass = type(name, (FileCatalogBase,),{"__init__": __init__, "__doc__":__doc__})
    return newclass

CSVCatalog       = FileCatalogFactory("CSVCatalog", io.CSVFile, examples='csv-data')
BinaryCatalog    = FileCatalogFactory("BinaryCatalog", io.BinaryFile, examples='binary-data')
BigFileCatalog   = FileCatalogFactory("BigFileCatalog", io.BigFile, examples='bigfile-data')
HDFCatalog       = FileCatalogFactory("HDFCatalog", io.HDFFile, examples='hdf-data')
TPMBinaryCatalog = FileCatalogFactory("TPMBinaryCatalog", io.TPMBinaryFile)
FITSCatalog      = FileCatalogFactory("FITSCatalog", io.FITSFile, examples='fits-data')
Gadget1Catalog   = FileCatalogFactory("Gadget1Catalog", io.Gadget1File, examples=None)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import set_options
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose
import pytest

# debug logging
setup_logging("debug")

@MPITest([1])
def test_tsc_interlacing(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)

    # interlacing with TSC
    mesh = source.to_mesh(resampler='tsc', Nmesh=64, interlaced=True, compensated=True)

    # compute the power spectrum -- should be flat shot noise
    # if the compensation worked
    r = FFTPower(mesh, mode='1d', kmin=0.02)
    # skip a few large scale modes that are noisier (fewer modes)
    assert_allclose(r.power['power'][5:], 1 / (3e-4), rtol=1e-1)

@MPITest([1])
def test_paint_empty(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)

    source = source[:0]
    assert source.csize == 0

    # interlacing with TSC
    mesh = source.to_mesh(resampler='tsc', Nmesh=64, interlaced=True, compensated=True)

    # compute the power spectrum -- should be flat shot noise
    # if the compensation worked
    real = mesh.to_real_field(normalize=True)
    assert_allclose(real, 1.0)

    real = mesh.to_real_field(normalize=False)
    assert_allclose(real, 0.0)

@MPITest([1])
def test_paint_chunksize(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)

    # interlacing with TSC
    mesh = source.to_mesh(resampler='tsc', Nmesh=64, interlaced=True, compensated=True)

    with set_options(paint_chunk_size=source.csize // 4):
        r1 = mesh.compute()

    with set_options(paint_chunk_size=source.csize):
        r2 = mesh.compute()

    assert_allclose(r1, r2)

@MPITest([1, 4])
def test_shotnoise(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)
    source['Weight'] = source.rng.uniform()

    # interlacing with TSC
    mesh = source.to_mesh(resampler='tsc', Nmesh=64, interlaced=True, compensated=True, weight='Weight')

    with set_options(paint_chunk_size=source.csize // 4):
        r1 = mesh.compute()

    with set_options(paint_chunk_size=source.csize):
        r2 = mesh.compute()
    assert_allclose(r1, r2)

    # expected shotnoise for uniform weights between 0 and 1
    SN = 4 / 3.0 * 1 / (3e-4)
    assert_allclose(r1.attrs['shotnoise'], SN, rtol=1e-2)
    assert_allclose(r2.attrs['shotnoise'], SN, rtol=1e-2)

@MPITest([1])
def test_cic_interlacing(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)

    # interlacing with TSC
    mesh = source.to_mesh(resampler='cic', Nmesh=64, interlaced=True, compensated=True)

    # compute the power spectrum -- should be flat shot noise
    # if the compensation worked
    r = FFTPower(mesh, mode='1d', kmin=0.02)

@MPITest([1])
def test_setters(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)

    # make the mesh
    mesh = source.to_mesh(resampler='cic', Nmesh=64, interlaced=True, compensated=True)

    assert mesh.compensated == True
    mesh.compensated = False
    assert mesh.compensated == False

    assert mesh.interlaced == True
    mesh.interlaced = False
    assert mesh.interlaced == False

    assert mesh.window == 'cic'
    mesh.window = 'tsc'
    assert mesh.window == 'tsc'

@MPITest([1])
def test_bad_window(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)

    # make the mesh
    mesh = source.to_mesh(resampler='cic', Nmesh=64, interlaced=True, compensated=True)

    # no such window
    with pytest.raises(Exception):
        mesh.window = "BAD"

@MPITest([1])
def test_no_compensation(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)

    # make the mesh
    mesh = source.to_mesh(resampler='cic', Nmesh=64, interlaced=True, compensated=True)

    # no compensation for this window
    mesh.window = 'db6'

    # cannot compute compensation
    with pytest.raises(ValueError):
        actions = mesh.actions

@MPITest([4])
def test_odd_chunksize(comm):
    # no errors shall occur. This is a regression test.

    source = ArrayCatalog({
        'Position': numpy.ones((2000, 3)),
    }, BoxSize=512., comm=comm)

    # make the mesh
    mesh = source.to_mesh(resampler='cic', Nmesh=64, interlaced=True, compensated=True)
    
    with set_options(paint_chunk_size=1111):
        mesh.compute()

@MPITest([1, 4])
def test_view(comm):

    # the CatalogSource
    source = UniformCatalog(nbar=2e-4, BoxSize=512., seed=42, comm=comm)
    source['TEST'] = 10.
    source.attrs['TEST'] = 10.0

    # the mesh
    mesh = source.to_mesh(Nmesh=32)

    # view
    view = mesh.view()
    assert view.base is mesh
    from nbodykit.base.mesh import MeshSource
    assert isinstance(view, MeshSource)

    # check meta-data
    for k in mesh.attrs:
        assert k in view.attrs

@MPITest([1, 4])
def test_apply_nocompensation(comm):

    # the CatalogSource
    source = UniformCatalog(nbar=2e-4, BoxSize=512, seed=42, comm=comm)
    source['TEST'] = 10.
    source['Position2'] = source['Position']
    source.attrs['TEST'] = 10.0

    # the mesh
    mesh = source.to_mesh(position='Position2', Nmesh=32, compensated=False)

    def raisefunc(k, v):
        raise StopIteration

    mesh = mesh.apply(raisefunc)

    with pytest.raises(StopIteration):
        mesh.compute()

    # view
    view = mesh.view()
    assert view.base is mesh
    assert isinstance(view, mesh.__class__)

    # check meta-data
    for k in mesh.attrs:
        assert k in view.attrs

@MPITest([1])
def test_apply_compensated(comm):

    # the CatalogSource
    source = UniformCatalog(nbar=2e-4, BoxSize=512., seed=42, comm=comm)
    source['TEST'] = 10.
    source['Position2'] = source['Position']
    source.attrs['TEST'] = 10.0

    # the mesh
    mesh = source.to_mesh(position='Position2', Nmesh=32, compensated=True)

    def raisefunc(k, v):
        raise StopIteration

    mesh = mesh.apply(raisefunc)

    with pytest.raises(StopIteration):
        mesh.compute()




from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

import pytest
import tempfile
import shutil
from numpy.testing import assert_allclose, assert_array_equal

setup_logging()

@MPITest([1,4])
def test_lost_attrs(comm):

    cosmo = cosmology.Planck15

    # initialize an output directory
    if comm.rank == 0:
        tmpfile = tempfile.mkdtemp()
    else:
        tmpfile = None
    tmpfile = comm.bcast(tmpfile)

    # linear mesh
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)

    # a hard to save attribute
    source.attrs['bad'] = cosmo.to_astropy()

    # generate a warning due to lost attr
    with pytest.warns(UserWarning):
        source.save(tmpfile, mode='real')

    # cleanup
    comm.barrier()
    if comm.rank == 0:
        shutil.rmtree(tmpfile)

@MPITest([1,4])
def test_real_save(comm):

    cosmo = cosmology.Planck15

    # initialize an output directory
    if comm.rank == 0:
        tmpfile = tempfile.mkdtemp()
    else:
        tmpfile = None
    tmpfile = comm.bcast(tmpfile)

    # linear mesh
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)

    # a hard to save attribute
    source.attrs['empty'] = None

    # save to bigfile
    source.save(tmpfile, mode='complex')

    # load as a BigFileMesh
    source2 = BigFileMesh(tmpfile, dataset='Field', comm=comm)

    # check sources
    for k in source.attrs:
        assert_array_equal(source2.attrs[k], source.attrs[k])

    # check data
    assert_array_equal(source2.compute(mode='complex'), source.compute(mode='complex'))

    # cleanup
    comm.barrier()
    if comm.rank == 0:
        shutil.rmtree(tmpfile)

@MPITest([1,4])
def test_real_save(comm):

    cosmo = cosmology.Planck15

    # initialize an output directory
    if comm.rank == 0:
        tmpfile = tempfile.mkdtemp()
    else:
        tmpfile = None
    tmpfile = comm.bcast(tmpfile)

    # linear mesh
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)

    # a hard to save attribute
    source.attrs['empty'] = None

    # save to bigfile
    source.save(tmpfile, mode='real')

    # load as a BigFileMesh
    source2 = BigFileMesh(tmpfile, dataset='Field', comm=comm)

    # check sources
    for k in source.attrs:
        assert_array_equal(source2.attrs[k], source.attrs[k])

    # check data
    assert_array_equal(source2.compute(mode='real'), source.compute(mode='real'))

    # cleanup
    comm.barrier()
    if comm.rank == 0:
        shutil.rmtree(tmpfile)

@MPITest([1,4])
def test_preview(comm):

    cosmo = cosmology.Planck15

    # linear mesh
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)

    # the painted RealField
    real = source.compute(mode='real')

    preview = source.preview()
    assert_allclose(preview.sum(), real.csum(), rtol=1e-5)

    real[...] **= 2
    preview[...] **= 2
    assert_allclose(preview.sum(), real.csum(), rtol=1e-5)

@MPITest([1,4])
def test_resample(comm):

    cosmo = cosmology.Planck15

    # linear mesh
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)

    # re-sample to Nmesh=32
    real = source.compute(mode='real', Nmesh=32)

    # and preview at same resolution
    preview = source.preview(Nmesh=32)
    assert_allclose(preview.sum(), real.csum(), rtol=1e-5)

    # XXX: disabled because currently paint uses fourier space resample
    # and preview uses configuration resample.

    #real[...] **= 2
    #preview[...] **= 2
    #assert_allclose(preview.sum(), real.csum(), rtol=1e-5)

@MPITest([1])
def test_bad_mode(comm):

    cosmo = cosmology.Planck15

    # linear mesh
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)

    with pytest.raises(ValueError):
        field = source.to_field(mode='BAD')

    with pytest.raises(ValueError):
        field = source.compute(mode='BAD')

@MPITest([1])
def test_meshfilter(comm):
    from nbodykit.base.mesh import MeshFilter
    class MyFilter(MeshFilter):
        kind = "wavenumber"
        mode = "complex"
        def filter(self, k, v):
            return 2.0 * v

    cosmo = cosmology.Planck15

    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)
    source2 = source.apply(MyFilter)
    source3 = source.apply(MyFilter())
    r1 = source.compute()
    r2 = source2.compute()
    r3 = source3.compute()
    assert_allclose(r1 * 2, r2)
    assert_allclose(r1 * 2, r3)

@MPITest([4])
def test_view(comm):

    from nbodykit.base.mesh import MeshSource
    cosmo = cosmology.Planck15

    # linear mesh
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)
    source.attrs['TEST'] = 10.0

    # view
    view = source.view()
    assert view.base is source
    assert isinstance(view, MeshSource)

    # check meta-data
    for k in source.attrs:
        assert k in view.attrs



from runtests.mpi import MPITest
from nbodykit.io.bigfile import BigFile
import shutil
import numpy
import tempfile
import pickle
import contextlib

@contextlib.contextmanager
def temporary_data():
    
    import bigfile
    try:
        data = numpy.empty(1024, dtype=[('Position', ('f8', 3)), ('Velocity', ('f8',3))])
        data['Position'] = numpy.random.random(size=(1024, 3))
        data['Velocity'] = numpy.random.random(size=(1024,3))

        tmpdir = tempfile.mkdtemp()
        with bigfile.File(tmpdir, create=True) as tmpff:
            with tmpff.create("Position", dtype=('f4', 3), size=1024) as bb:
                bb.write(0, data['Position'])
            with tmpff.create("Velocity", dtype=('f4', 3), size=1024) as bb:
                bb.write(0, data['Velocity'])
            with tmpff.create("Header") as bb:
                bb.attrs['Size'] = 1024
                bb.attrs['Over'] = 0

            with tmpff.create('1/.') as bb:
                bb.attrs['Size'] = 1024
                bb.attrs['Over'] = 1024

            with tmpff.create("1/Position", dtype=('f4', 3), size=1024) as bb:
                bb.write(0, data['Position'])
            with tmpff.create("1/Velocity", dtype=('f4', 3), size=1024) as bb:
                bb.write(0, data['Velocity'])

        yield (data, tmpdir)
    except:
        raise
    finally:
        shutil.rmtree(tmpdir)

@MPITest([1])
def test_data(comm):

    with temporary_data() as (data, tmpfile):
        # read
        ff = BigFile(tmpfile, header='Header', dataset='1')

        # check size
        assert ff.attrs['Size'] == 1024

        # and data
        numpy.testing.assert_almost_equal(data['Position'], ff['Position'][:])
        numpy.testing.assert_almost_equal(data['Velocity'], ff['Velocity'][:])

        numpy.testing.assert_almost_equal(data['Velocity'][::2], ff['Velocity'][::2])

@MPITest([1])
def test_data_auto_header(comm):
    with temporary_data() as (data, tmpfile):
        ff = BigFile(tmpfile, dataset='1')

        # check size
        assert ff.attrs['Size'] == 1024
        assert ff.attrs['Over'] == 1024

@MPITest([1])
def test_pickle(comm):
    
    with temporary_data() as (data, tmpfile):
        
        # read
        ff = BigFile(tmpfile, header='Header', exclude=['1/*', 'Header'])
    
        # pickle
        s = pickle.dumps(ff)
        ff2 = pickle.loads(s)

        # check size
        assert ff2.attrs['Size'] == 1024
        
        # and data
        numpy.testing.assert_almost_equal(data['Position'], ff2['Position'][:])
        numpy.testing.assert_almost_equal(data['Velocity'], ff2['Velocity'][:])




from .bigfile import BigFile
from .binary import BinaryFile
from .csv import CSVFile
from .hdf import HDFFile
from .stack import FileStack
from .tpm import TPMBinaryFile
from .gadget import Gadget1File
from .fits import FITSFile



from .bigfile import BigFileMesh
from .linear import LinearMesh
from .field import FieldMesh
from .array import ArrayMesh

from .species import MultipleSpeciesCatalogMesh
from .catalog import CatalogMesh

__all__ = ['BigFileMesh',
           'LinearMesh',
           'FieldMesh',
           'ArrayMesh',
           'CatalogMesh',
           'MultipleSpeciesCatalogMesh',
          ]



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
import pytest

setup_logging()

@MPITest([4])
def test_bad_init(comm):

    # initialize a catalog
    cat = UniformCatalog(nbar=100, BoxSize=1.0, comm=comm)
    cat['Mass'] = 1.0

    # cannot specify column as None
    with pytest.raises(ValueError):
        halos = HaloCatalog(cat, cosmology.Planck15, 0., mass=None)

    # missing column
    with pytest.raises(ValueError):
        halos = HaloCatalog(cat, cosmology.Planck15, 0., mass='MISSING')

@MPITest([4])
def test_missing_boxsize(comm):

    # initialize a catalog
    cat = UniformCatalog(nbar=100, BoxSize=1.0, comm=comm)
    cat['Mass'] = 1.0

    # initialize halos
    halos = HaloCatalog(cat, cosmology.Planck15, 0.)

    # delete BoxSize
    del halos.attrs['BoxSize']

    # missing BoxSize!
    with pytest.raises(ValueError):
        halocat = halos.to_halotools()



import os
import numpy
import logging

from nbodykit.utils import split_size_3d
from pmesh.domain import GridND
from scipy.spatial.ckdtree import cKDTree as KDTree

class KDDensity(object):
    """
    Estimate a proxy density based on the distance to the nearest neighbor.
    The result is proportional to the density but the scale is unspecified.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~KDDensity.run` for the attributes storing the results.

    Parameters
    ----------
    source : CatalogSource
        the input source of particles to compute the proxy density on;
        must specify the 'Position' column
    margin: float, optional
        Padding region per parallel domain; relative to the mean seperation
    """
    logger = logging.getLogger('KDDensity')

    def __init__(self, source, margin=1.0):

        if 'Position' not in source:
            raise ValueError("please specify the 'Position' column in the input source")

        self.comm = source.comm
        self._source = source

        if 'BoxSize' not in self._source.attrs:
            raise ValueError("please specify 'BoxSize' in the input source 'attrs'")
        BoxSize = numpy.array(self._source.attrs['BoxSize'], dtype='f8')
        if BoxSize.ndim == 0 or len(BoxSize) == 1: # catch some common problems about BoxSize.
            BoxSize = numpy.array([BoxSize, BoxSize, BoxSize])

        # store some meta-data
        self.attrs = {}
        self.attrs['BoxSize'] = BoxSize
        self.attrs['meansep'] = (self._source.csize / BoxSize.prod()) ** (1.0 / len(BoxSize))
        self.attrs['margin'] = margin

        # run the algorithm
        self.run()

    def run(self):
        """
        Compute the density proxy. This attaches the following attribute:

        - :attr:`density`

        Attributes
        ----------
        density : array_like, length: :attr:`size`
            a unit-less, proxy density value for each object on the local
            rank. This is computed as the inverse cube of the distance
            to the closest, nearest neighbor
        """

        # do the domain decomposition
        Np = split_size_3d(self.comm.size)
        edges = [numpy.linspace(0, self.attrs['BoxSize'][d], Np[d] + 1, endpoint=True) for d in range(3)]
        domain = GridND(comm=self.comm, periodic=True, edges=edges)

        # read all position and exchange
        pos = self._source.compute(self._source['Position'])
        layout = domain.decompose(pos, smoothing=self.attrs['margin'] * self.attrs['meansep'])
        xpos = layout.exchange(pos)

        # wait for scipy 0.19.1
        assert all(self.attrs['BoxSize'] == self.attrs['BoxSize'][0])
        xpos[...] /= self.attrs['BoxSize']
        xpos %= 1

        # KDTree
        tree = KDTree(xpos, boxsize=1.0)
        d, i = tree.query(xpos, k=[8])
        d = d[:, 0]

        # gather back to original root, taking the minimum distance
        d = layout.gather(d, mode=numpy.fmin)
        self.density = 1 / (d ** 3 * self.attrs['BoxSize'].prod())



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from scipy.interpolate import InterpolatedUnivariateSpline
from numpy.testing import assert_allclose, assert_array_equal
import pytest

setup_logging("debug")

NDATA = 1000
NBAR = 1e-4

def make_sources(cosmo, comm):

    data = RandomCatalog(NDATA, seed=42, comm=comm)
    randoms = RandomCatalog(NDATA*10, seed=84, comm=comm)

    # add the random columns
    for s in [data, randoms]:

        # ra, dec, z
        s['z']   = s.rng.normal(loc=0.5, scale=0.1)
        s['ra']  = s.rng.uniform(low=110, high=260)
        s['dec'] = s.rng.uniform(low=-3.6, high=60.)

        # position
        s['Position'] = transform.SkyToCartesian(s['ra'], s['dec'], s['z'], cosmo=cosmo)

    return data, randoms

@MPITest([1])
def test_add_fkpweight(comm):
    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo, comm)
    for s in [data, randoms]:
        # constant number density
        s['NZ'] = NBAR

    P0 = 1e4

    # two mesh objects from same FKP source
    fkp = FKPCatalog(data, randoms, P0=P0)

    assert_allclose(fkp['data']['FKPWeight'].compute(),
                    FKPWeightFromNbar(P0, fkp['data']['NZ'].compute()))

    assert_allclose(fkp['data']['FKPWeight'].compute(),
                    FKPWeightFromNbar(P0, fkp['data']['NZ'].compute()))

    # updating NZ shall affect FKPWeight if it is the default.
    for s in [data, randoms]:
        # constant number density
        s['NZ'] = 2 * NBAR

    assert_allclose(fkp['data']['FKPWeight'].compute(),
                    FKPWeightFromNbar(P0, fkp['data']['NZ'].compute()))

    assert_allclose(fkp['data']['FKPWeight'].compute(),
                    FKPWeightFromNbar(P0, fkp['data']['NZ'].compute()))

@MPITest([1, 4])
def test_diff_cross_boxsizes(comm):

    cosmo = cosmology.Planck15
    P0 = 1e4

    # make the sources
    data, randoms = make_sources(cosmo, comm)
    for s in [data, randoms]:

        # constant number density
        s['NZ'] = NBAR

        # completeness weights
        s['Weight'] = (1 + P0*s['NZ'])**2


    # two mesh objects from same FKP source
    fkp = FKPCatalog(data, randoms, P0=P0)
    mesh1 = fkp.to_mesh(Nmesh=128, dtype='f8')

    # second mesh has larger box size
    mesh2 = fkp.to_mesh(Nmesh=128, dtype='f8', BoxSize=1.1*mesh1.attrs['BoxSize'])

    # compute the multipoles
    r = ConvolvedFFTPower(mesh1, second=mesh2, poles=[0,2,4], dk=0.005)

    # make sure we matched the box sizes
    assert_array_equal(r.first.attrs['BoxSize'], r.second.attrs['BoxSize'])
    assert_array_equal(r.first.attrs['BoxCenter'], r.second.attrs['BoxCenter'])

@MPITest([1, 4])
def test_true_cross_corr_fail(comm):

    cosmo = cosmology.Planck15
    P0 = 1e4

    # make the sources
    data, randoms = make_sources(cosmo, comm)
    for s in [data, randoms]:
        s['NZ'] = NBAR

    # two mesh objects from different FKP source
    fkp1 = FKPCatalog(data, randoms, P0=P0)
    fkp2 = FKPCatalog(data, randoms, P0=P0)

    mesh1 = fkp1.to_mesh(Nmesh=128, dtype='f8')
    mesh2 = fkp2.to_mesh(Nmesh=128, dtype='f8')

    # cannot do cross correlations with different data/randoms catalogs
    with pytest.raises(NotImplementedError):
        r = ConvolvedFFTPower(mesh1, second=mesh2, poles=[0,2,4], dk=0.005)


@MPITest([1, 4])
def test_bad_cross_corr_columns(comm):

    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo, comm)

    for s in [data, randoms]:
        s['NZ'] = NBAR

    # same FKP source but different selection columns won't work!
    fkp = FKPCatalog(data, randoms)
    fkp['data']['Selection2'] = fkp['data/Selection']
    fkp['randoms']['Selection2'] = fkp['randoms/Selection']

    mesh1 = fkp.to_mesh(Nmesh=128, dtype='f8', selection='Selection')
    mesh2 = fkp.to_mesh(Nmesh=128, dtype='f8', selection='Selection2')

    # columns in mesh must be same except for weight
    # YF: FIXME: why this limitation?
    with pytest.raises(NotImplementedError):
        r = ConvolvedFFTPower(mesh1, second=mesh2, poles=[0,2,4], dk=0.005)

@MPITest([1, 4])
def test_cross_corr(comm):

    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo, comm)
    P0 = 1e4

    for s in [data, randoms]:

        # constant number density
        s['NZ'] = NBAR

        # completeness weights
        s['Weight'] = (1 + P0*s['NZ'])**2

    # two mesh objects from same FKP source
    fkp = FKPCatalog(data, randoms, P0=P0)
    mesh1 = fkp.to_mesh(Nmesh=128, dtype='f8', comp_weight='Weight', selection='Selection')
    mesh2 = fkp.to_mesh(Nmesh=128, dtype='f8', comp_weight='Weight', selection='Selection')

    # compute the multipoles
    r = ConvolvedFFTPower(mesh1, second=mesh2, poles=[0,2,4], dk=0.005)

    # normalization
    assert_allclose(r.attrs['data.norm'], NDATA*NBAR)
    assert_allclose(r.attrs['randoms.norm'], NDATA*NBAR)

    # shotnoise
    S_data = r.attrs['first.data.W']/r.attrs['randoms.norm']
    S_ran = r.attrs['first.randoms.W']/r.attrs['randoms.norm']*r.attrs['alpha']**2
    S = S_data + S_ran
    assert_allclose(S, r.attrs['shotnoise'])

@MPITest([1, 4])
def test_bad_input(comm):

    cosmo = cosmology.Planck15

    # source has wrong type
    data, randoms = make_sources(cosmo, comm)
    for s in [data, randoms]:
        s['NZ'] = NBAR
        s['FKPWeight'] = 1.0 / (1 + 2e4*s['NZ'])

    with pytest.raises(TypeError):
        r = ConvolvedFFTPower(data, poles=[0,2,4], dk=0.005, Nmesh=64)

    # the FKP source
    fkp = FKPCatalog(data, randoms)

    for s in [data, randoms]:
        assert_allclose(s['FKPWeight'], 1.0 / (1 + 2e4*s['NZ']))

    # must specify P0_FKP
    with pytest.raises(ValueError):
        r = ConvolvedFFTPower(fkp, poles=0, dk=0.005, use_fkp_weights=True, P0_FKP=None, Nmesh=64)

    # warn about overwriting FKP Weights
    with pytest.raises(ValueError):
        r = ConvolvedFFTPower(fkp, poles=0, dk=0.005, use_fkp_weights=True, P0_FKP=1e4, Nmesh=64)


@MPITest([4])
def test_no_monopole(comm):

    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo, comm)

    # select in given redshift range
    for s in [data, randoms]:
        s['NZ'] = NBAR

    # the FKP source
    fkp = FKPCatalog(data, randoms, nbar='NZ')
    fkp = fkp.to_mesh(Nmesh=128, dtype='f8')

    # compute the multipoles
    r = ConvolvedFFTPower(fkp, poles=[2,4], dk=0.005)

    assert 'power_0' not in r.poles.variables
    assert 'power_2' in r.poles.variables
    assert 'power_4' in r.poles.variables

@MPITest([4])
def test_bad_normalization(comm):

    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo, comm)

    # select in given redshift range
    for s in [data, randoms]:
        s['NZ'] = NBAR

    # bad normalization via NZ
    randoms['NZ'] *= 50.0

    # the FKP source
    fkp = FKPCatalog(data, randoms, nbar='NZ')
    fkp = fkp.to_mesh(Nmesh=128, dtype='f8')

    # compute the multipoles
    with pytest.raises(ValueError):
        r = ConvolvedFFTPower(fkp, poles=[0,2,4], dk=0.005)

@MPITest([4])
def test_selection(comm):

    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo, comm)

    # select in given redshift range
    for s in [data, randoms]:
        s['NZ'] = NBAR
        s['Selection'] = (s['z'] > 0.4)&(s['z'] < 0.6)

    # the FKP source
    fkp = FKPCatalog(data, randoms, nbar='NZ')
    fkp = fkp.to_mesh(Nmesh=128, dtype='f8', selection='Selection')

    # compute the multipoles
    r = ConvolvedFFTPower(fkp, poles=[0,2,4], dk=0.005)

    # number of data objects selected
    N = comm.allreduce(((data['z'] > 0.4)&(data['z'] < 0.6)).sum())
    assert_allclose(r.attrs['data.N'], N)

    # number of randoms selected
    N = comm.allreduce(((randoms['z'] > 0.4)&(randoms['z'] < 0.6)).sum())
    assert_allclose(r.attrs['randoms.N'], N)

    # and save
    r.save("conv-power-with-selection.json")

    # load and check output
    r2 = ConvolvedFFTPower.load("conv-power-with-selection.json", comm=comm)
    assert_array_equal(r.poles.data, r2.poles.data)

@MPITest([1, 4])
def test_run(comm):

    cosmo = cosmology.Planck15
    P0 = 1e4

    # make the sources
    data, randoms = make_sources(cosmo, comm)
    for s in [data, randoms]:

        # constant number density
        s['NZ'] = NBAR

        # completeness weights
        s['Weight'] = (1 + P0*s['NZ'])**2

    # the FKP source
    fkp = FKPCatalog(data, randoms, P0=P0, nbar='NZ')
    fkp = fkp.to_mesh(Nmesh=128, dtype='f8', fkp_weight='FKPWeight', comp_weight='Weight', selection='Selection')

    # compute the multipoles
    r = ConvolvedFFTPower(fkp, poles=[0,2,4], dk=0.005)

    # compute pkmu
    mu_edges = numpy.linspace(0, 1, 6)
    pkmu = r.to_pkmu(mu_edges=mu_edges, max_ell=4)

    # normalization
    assert_allclose(r.attrs['data.norm'], NDATA*NBAR)
    assert_allclose(r.attrs['randoms.norm'], NDATA*NBAR)

    # shotnoise
    S_data = r.attrs['data.W']/r.attrs['randoms.norm']
    S_ran = r.attrs['randoms.W']/r.attrs['randoms.norm']*r.attrs['alpha']**2
    S = S_data + S_ran
    assert_allclose(S, r.attrs['shotnoise'])

@MPITest([1, 4])
def test_run_unique_bins(comm):

    cosmo = cosmology.Planck15
    P0 = 1e4

    # make the sources
    data, randoms = make_sources(cosmo, comm)
    for s in [data, randoms]:

        # constant number density
        s['NZ'] = NBAR

        # completeness weights
        s['Weight'] = (1 + P0*s['NZ'])**2

    # the FKP source
    fkp = FKPCatalog(data, randoms, P0=P0, nbar='NZ')
    fkp = fkp.to_mesh(Nmesh=128, dtype='f8', fkp_weight='FKPWeight', comp_weight='Weight', selection='Selection')

    # compute the multipoles
    r = ConvolvedFFTPower(fkp, poles=[0,2,4], dk=0)

    # compute pkmu
    mu_edges = numpy.linspace(0, 1, 6)
    pkmu = r.to_pkmu(mu_edges=mu_edges, max_ell=4)
    assert_allclose(pkmu.coords['k'], r.poles.coords['k'])

@MPITest([1, 4])
def test_run_unique_bins_windowonly(comm):

    cosmo = cosmology.Planck15
    P0 = 1e4

    # make the sources
    data, randoms = make_sources(cosmo, comm)
    for s in [data, randoms]:

        # constant number density
        s['NZ'] = NBAR

        # completeness weights
        s['Weight'] = (1 + P0*s['NZ'])**2

    # the FKP source
    fkp = FKPCatalog(data=randoms, randoms=None, P0=P0, nbar='NZ')
    fkp = fkp.to_mesh(Nmesh=128, dtype='f8', fkp_weight='FKPWeight', comp_weight='Weight', selection='Selection')

    # compute the multipoles
    r = ConvolvedFFTPower(fkp, poles=[0,2,4], dk=0)

    # compute pkmu
    mu_edges = numpy.linspace(0, 1, 6)
    pkmu = r.to_pkmu(mu_edges=mu_edges, max_ell=4)
    assert_allclose(pkmu.coords['k'], r.poles.coords['k'])

@MPITest([1, 4])
def test_window_only(comm):
    NDATA = 1000
    NBAR = 1e-4
    FSKY = 0.15
    P0 = 1e4

    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo, comm)

    for s in [data, randoms]:

        # constant number density
        s['NZ'] = NBAR

        # completeness weights
        s['Weight'] = (1 + P0*s['NZ'])**2

    empty = randoms[:0]
    # initialize the FKP source with random as data, for measuring the window.
    fkp = FKPCatalog(data=randoms, randoms=None, P0=P0)

    # compute the multipoles
    r = ConvolvedFFTPower(fkp.to_mesh(Nmesh=128), poles=[0,2,4], dk=0.005)

    assert not numpy.isnan(r.poles['power_0']).any()
    assert not numpy.isnan(r.poles['power_2']).any()
    assert not numpy.isnan(r.poles['power_4']).any()

@MPITest([1, 4])
def test_with_zhist(comm):

    NDATA = 1000
    NBAR = 1e-4
    FSKY = 0.15

    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo, comm)

    # compute NZ from randoms
    zhist = RedshiftHistogram(randoms, FSKY, cosmo, redshift='z')

    # normalize NZ to the total size of the data catalog
    alpha = 1.0 * data.csize / randoms.csize
    # add n(z) from randoms to the FKP source

    randoms['NZ'] = zhist.interpolate(randoms['z']) * alpha
    data['NZ'] = zhist.interpolate(data['z']) * alpha

    # initialize the FKP source
    fkp = FKPCatalog(data, randoms)

    # compute the multipoles
    r = ConvolvedFFTPower(fkp.to_mesh(Nmesh=128), poles=[0,2,4], dk=0.005)

    assert_allclose(r.attrs['data.norm'], 0.000388338522187, rtol=1e-4)
    assert_allclose(r.attrs['randoms.norm'], 0.000395808747269, rtol=1e-4)



from nbodykit.source.mesh.catalog import CatalogMesh
from nbodykit import _global_options
import numpy
import logging
import warnings

# for converting from particle to mesh
from pmesh.pm import RealField

from nbodykit.utils import attrs_to_dict
class MultipleSpeciesCatalogMesh(CatalogMesh):
    """
    A subclass of :class:`~nbodykit.base.catalogmesh.CatalogMesh`
    designed to paint the density field from a sum of multiple types
    of particles.

    The :func:`compute` function paints the density field summed over
    all particle species.

    Parameters
    ----------
    source : CatalogSource
        the input catalog that we wish to interpolate to a mesh
    BoxSize :
        the size of the box
    Nmesh : int, 3-vector
        the number of cells per mesh side
    dtype : str
        the data type of the values stored on mesh
    weight : str
        column in ``source`` that specifies the weight value for each
        particle in the ``source`` to use when gridding
    value : str
        column in ``source`` that specifies the field value for each particle;
        the mesh stores a weighted average of this column
    selection : str
        column in ``source`` that selects the subset of particles to grid
    position : str, optional
        column in ``source`` specifying the position coordinates; default
        is ``Position``
    """
    logger = logging.getLogger('MultipleSpeciesCatalogMesh')

    def __init__(self, source, Nmesh, BoxSize, dtype,
            selection, position, weight, value,
            interlaced, compensated, resampler):

        from nbodykit.source.catalog import MultipleSpeciesCatalog

        if not isinstance(source, MultipleSpeciesCatalog):
            raise TypeError(("the input source for MultipleSpeciesCatalogMesh "
                             "must be a MultipleSpeciesCatalog"))

        CatalogMesh.__init__(self,
                    source=source,
                    Nmesh=Nmesh, BoxSize=BoxSize, dtype=dtype,
                    Position=None,
                    Selection=None,
                    Weight=None,
                    Value=None,
                    resampler=resampler,
                    compensated=compensated,
                    interlaced=interlaced,
                    comm=source.comm)

        self.species = source.species

        self.position = position
        self.selection = selection
        self.weight = weight
        self.value = value

    def __iter__(self):
        return iter(self.species)

    def __getitem__(self, key):
        """
        If indexed by a species name, return a CatalogMesh object holding
        only the data columns for that species with the same parameters as
        the current object.

        If not a species name, this has the same behavior as
        :func:`CatalogSource.__getitem__`.
        """
        from nbodykit.source.mesh import CatalogMesh

        if key in self.source.species:

            cat = self.source[key]

            # view as a catalog mesh
            mesh = CatalogMesh(cat,
                    BoxSize=self.attrs['BoxSize'],
                    Nmesh=self.attrs['Nmesh'],
                    dtype=self.dtype,
                    Weight=cat[self.weight],
                    Value=cat[self.value],
                    Selection=cat[self.selection],
                    Position=cat[self.position],
                    interlaced=self.interlaced,
                    compensated=self.compensated,
                    resampler=self.resampler,
                )

            # attach attributes from self
            return mesh.__finalize__(self)
        else:
            raise KeyError("%s is not a species defined in the source" % key)

    def to_real_field(self, normalize=True):
        r"""
        Paint the density field holding the sum of all particle species,
        returning a :class:`~pmesh.pm.RealField` object.

        Meta-data computed for each particle is stored in the :attr:`attrs`
        attribute of the returned RealField, with keys that are prefixed by
        the species name. In particular, the total shot noise for the
        mesh is defined as:

        .. math::

            P_\mathrm{shot} = \sum_i (W_i/W_\mathrm{tot})^2 P_{\mathrm{shot},i},

        where the sum is over all species in the catalog, ``W_i`` is the
        sum of the ``Weight`` column for the :math:`i^\mathrm{th}` species,
        and :math:`W_\mathrm{tot}` is the sum of :math:`W_i` across all species.

        Parameters
        ----------
        normalize : bool, optional
            if ``True``, normalize the density field as :math:`1+\delta`,
            dividing by the total mean number of objects per cell, as given
            by the ``num_per_cell`` meta-data value in :attr:`attrs`

        Returns
        -------
        RealField :
            the RealField holding the painted density field, with a
            :attr:`attrs` dictionary attribute holding the meta-data
        """
        # track the sum of the mean number of objects per cell across species
        attrs = {'num_per_cell':0., 'N':0}

        # initialize an empty real field
        real = self.pm.create(type='real', value=0)

        # loop over each species
        for name in self.source.species:

            if self.pm.comm.rank == 0:
                self.logger.info("painting the '%s' species" %name)

            # get a CatalogMesh for this species
            species_mesh = self[name]

            # paint (in-place) the un-normalized density field for this species
            species_mesh.to_real_field(out=real, normalize=False)

            # add to the mean number of objects per cell and total number
            attrs['num_per_cell'] += real.attrs['num_per_cell']
            attrs['N'] += real.attrs['N']

            # store the meta-data for this species, with a prefix
            attrs.update(attrs_to_dict(real, name+'.'))

        # # normalize the field by nbar -> this is now 1+delta
        if normalize:
            real[:] /= attrs['num_per_cell']

        # update the meta-data
        real.attrs.clear()
        real.attrs.update(attrs)

        # compute total shot noise by summing of shot noise each species
        real.attrs['shotnoise'] = 0
        total_weight = sum(real.attrs['%s.W' %name] for name in self.source.species)
        for name in self.source.species:
            this_Pshot = real.attrs['%s.shotnoise' %name]
            this_weight = real.attrs['%s.W' %name]
            real.attrs['shotnoise'] += (this_weight/total_weight)**2 * this_Pshot

        return real



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose

from nbodykit.algorithms.kdtree import KDDensity

# debug logging
setup_logging("debug")

@MPITest([1, 4])
def test_kddensity(comm):
    cosmo = cosmology.Planck15

    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-4, BoxSize=64., Nmesh=16, seed=42, comm=comm)

    kdden = KDDensity(source)
    assert kdden.density.size == source.size
    print(kdden.density.max())



from .cosmology import Cosmology
from .background import PerturbationGrowth
from .power import *
from .correlation import CorrelationFunction, xi_to_pk, pk_to_xi

# override these with Cosmology classes below
from astropy.cosmology import Planck13, Planck15, WMAP5, WMAP7, WMAP9

# Planck defaults with sigma8, n_s
_kws = {'ln10^{10}A_s':3.0973, 'n_s':0.9611, 'k_pivot':0.05, 'tau_reio':0.0952}
Planck13 = Cosmology.from_astropy(Planck13, **_kws)
"""Planck13 instance of FlatLambdaCDM cosmology

From Planck Collaboration 2014, A&A, 571, A16 (Paper XVI), Table 5 (Planck + WP + highL + BAO)
"""

_kws = {'ln10^{10}A_s':3.064, 'n_s':0.9667, 'k_pivot':0.05, 'tau_reio':0.066}
Planck15 = Cosmology.from_astropy(Planck15, **_kws)
"""Planck15 instance of FlatLambdaCDM cosmology

From Planck Collaboration 2016, A&A, 594, A13 (Paper XIII), Table 4 (TT, TE, EE + lowP + lensing + ext)
"""

# WMAP defaults with sigma8, n_s
_kws = {'A_s':2.46e-9, 'k_pivot':0.002, 'n_s':0.962, 'tau_reio':0.088}
WMAP5 = Cosmology.from_astropy(WMAP5, **_kws)
"""
WMAP5 instance of FlatLambdaCDM cosmology

From Komatsu et al. 2009, ApJS, 180, 330, doi: 10.1088/0067-0049/180/2/330.
Table 1 (WMAP + BAO + SN ML).
"""

_kws = {'A_s':2.42e-9, 'k_pivot':0.002, 'n_s':0.967, 'tau_reio':0.085}
WMAP7 = Cosmology.from_astropy(WMAP7, **_kws)
"""
WMAP7 instance of FlatLambdaCDM cosmology

From Komatsu et al. 2011, ApJS, 192, 18, doi: 10.1088/0067-0049/192/2/18.
Table 1 (WMAP + BAO + H0 ML).
"""

_kws = {'A_s':2.464e-9, 'k_pivot':0.002, 'n_s':0.9608, 'tau_reio':0.081}
WMAP9 = Cosmology.from_astropy(WMAP9, **_kws)
"""WMAP9 instance of FlatLambdaCDM cosmology

From Hinshaw et al. 2013, ApJS, 208, 19, doi: 10.1088/0067-0049/208/2/19.
Table 4 (WMAP9 + eCMB + BAO + H0, last column)
"""



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_allclose, assert_array_equal

setup_logging()

@MPITest([1,4])
def test_paint(comm):
    from pmesh.pm import ParticleMesh

    pm = ParticleMesh(Nmesh=(8, 8, 8), BoxSize=(128, 128, 128.), comm=comm)
    real = pm.generate_whitenoise(type='real', seed=3333)
    complex = real.r2c()

    realmesh = FieldMesh(real)
    complexmesh = FieldMesh(complex)

    assert_array_equal(real, realmesh.to_field(mode='real'))
    assert_array_equal(complex, complexmesh.to_field(mode='complex'))

    assert_allclose(complex, realmesh.to_field(mode='complex'))
    assert_allclose(real, complexmesh.to_field(mode='real'))



import numpy

available = ['CLASS', 'EisensteinHu', 'NoWiggleEisensteinHu']

# minimum CLASS value to represent k->0
KMIN = 1e-8

class CLASS(object):
    """
    The linear matter transfer function using the CLASS Boltzmann code.

    Parameters
    ----------
    cosmo : :class:`Cosmology`
        the cosmology instance
    redshift : float
        the redshift of the power spectrum
    """
    def __init__(self, cosmo, redshift):

        # make sure we have Pk
        if not cosmo.has_pk_matter:
            cosmo = cosmo.clone(output='mPk dTk vTk')
        self.cosmo = cosmo

        # find the low-k amplitude to normalize to 1 as k->0 at z = 0
        self._norm = 1.0; self.redshift = 0
        self._norm = 1. / self(KMIN)

        # the proper redshift
        self.redshift = redshift

    def __call__(self, k):
        r"""
        Return the CLASS linear transfer function at :attr:`redshift`. This
        computes the transfer function from the CLASS linear power spectrum
        using:

        .. math::

            T(k) = \left [ P_L(k) / k^n_s  \right]^{1/2}.

        We normalize the transfer function :math:`T(k)` to unity as
        :math:`k \rightarrow 0` at :math:`z=0`.

        Parameters
        ---------
        k : float, array_like
            the wavenumbers in units of :math:`h \mathrm{Mpc}^{-1}`

        Returns
        -------
        Tk : float, array_like
            the transfer function evaluated at ``k``, normalized to unity on
            large scales
        """
        k = numpy.asarray(k)
        nonzero = k>0
        linearP = self.cosmo.get_pklin(k[nonzero], self.redshift) / self.cosmo.h**3 # in Mpc^3
        primordialP = (k[nonzero]*self.cosmo.h)**self.cosmo.n_s # put k into Mpc^{-1}

        # return shape
        Tk = numpy.ones(nonzero.shape)

        # at k=0, this is 1.0 * D(z), where T(k) = 1.0 at z=0
        Tk[~nonzero] = self.cosmo.scale_independent_growth_factor(self.redshift)

        # fill in all k>0
        Tk[nonzero] = self._norm * (linearP / primordialP)**0.5
        return Tk


class EisensteinHu(object):
    """
    The linear matter transfer function using the Eisenstein & Hu (1998)
    fitting formula with BAO wiggles.

    Parameters
    ----------
    cosmo : :class:`Cosmology`
        the cosmology instance
    redshift : float
        the redshift of the power spectrum

    References
    ----------
    Eisenstein & Hu, "Baryonic Features in the Matter Transfer Function", 1998
    """
    def __init__(self, cosmo, redshift):

        self.cosmo = cosmo
        self.redshift = redshift

        self.Obh2 = cosmo.Omega0_b * cosmo.h ** 2
        self.Omh2 = cosmo.Omega0_m * cosmo.h ** 2
        self.f_baryon = cosmo.Omega0_b / cosmo.Omega0_m
        self.theta_cmb = cosmo.Tcmb0 / 2.7

        # redshift and wavenumber of equality
        self.z_eq = 2.5e4 * self.Omh2 * self.theta_cmb ** (-4) # this is 1 + z
        self.k_eq = 0.0746 * self.Omh2 * self.theta_cmb ** (-2) # units of 1/Mpc

        # sound horizon and k_silk
        self.z_drag_b1 = 0.313 * self.Omh2 ** -0.419 * (1 + 0.607 * self.Omh2 ** 0.674)
        self.z_drag_b2 = 0.238 * self.Omh2 ** 0.223
        self.z_drag    = 1291 * self.Omh2 ** 0.251 / (1. + 0.659 * self.Omh2 ** 0.828) * \
                           (1. + self.z_drag_b1 * self.Obh2 ** self.z_drag_b2)

        self.r_drag = 31.5 * self.Obh2 * self.theta_cmb ** -4 * (1000. / (1+self.z_drag))
        self.r_eq   = 31.5 * self.Obh2 * self.theta_cmb ** -4 * (1000. / self.z_eq)

        self.sound_horizon = 2. / (3.*self.k_eq) * numpy.sqrt(6. / self.r_eq) * \
                    numpy.log((numpy.sqrt(1 + self.r_drag) + numpy.sqrt(self.r_drag + self.r_eq)) / (1 + numpy.sqrt(self.r_eq)) )
        self.k_silk = 1.6 * self.Obh2 ** 0.52 * self.Omh2 ** 0.73 * (1 + (10.4*self.Omh2) ** -0.95)

        # alpha_c
        alpha_c_a1 = (46.9*self.Omh2) ** 0.670 * (1 + (32.1*self.Omh2) ** -0.532)
        alpha_c_a2 = (12.0*self.Omh2) ** 0.424 * (1 + (45.0*self.Omh2) ** -0.582)
        self.alpha_c = alpha_c_a1 ** -self.f_baryon * alpha_c_a2 ** (-self.f_baryon**3)

        # beta_c
        beta_c_b1 = 0.944 / (1 + (458*self.Omh2) ** -0.708)
        beta_c_b2 = 0.395 * self.Omh2 ** -0.0266
        self.beta_c = 1. / (1 + beta_c_b1 * ((1-self.f_baryon) ** beta_c_b2) - 1)

        y = self.z_eq / (1 + self.z_drag)
        alpha_b_G = y * (-6.*numpy.sqrt(1+y) + (2. + 3.*y) * numpy.log((numpy.sqrt(1+y)+1) / (numpy.sqrt(1+y)-1)))
        self.alpha_b = 2.07 *  self.k_eq * self.sound_horizon * (1+self.r_drag)**-0.75 * alpha_b_G

        self.beta_node = 8.41 * self.Omh2 ** 0.435
        self.beta_b    = 0.5 + self.f_baryon + (3. - 2.*self.f_baryon) * numpy.sqrt( (17.2*self.Omh2) ** 2 + 1 )

    def __call__(self, k):
        r"""
        Return the Eisenstein-Hu transfer function with BAO wiggles.

        This is normalized to unity as :math:`k \rightarrow 0` at :math:`z=0`.
        The redshift scaling is provided by the
        :func:`Cosmology.scale_independent_growth_factor` function.

        Parameters
        ---------
        k : float, array_like
            the wavenumbers in units of :math:`h \mathrm{Mpc}^{-1}`

        Returns
        -------
        Tk : float, array_like
            the transfer function evaluated at ``k``, normalized to unity on
            large scales
        """
        if numpy.isscalar(k) and k == 0.:
            return 1.0

        k = numpy.asarray(k)
        # only compute k > 0 modes
        valid = k > 0.

        k = k[valid] * self.cosmo.h # now in 1/Mpc
        q = k / (13.41*self.k_eq)
        ks = k*self.sound_horizon

        T_c_ln_beta   = numpy.log(numpy.e + 1.8*self.beta_c*q)
        T_c_ln_nobeta = numpy.log(numpy.e + 1.8*q);
        T_c_C_alpha   = 14.2 / self.alpha_c + 386. / (1 + 69.9 * q ** 1.08)
        T_c_C_noalpha = 14.2 + 386. / (1 + 69.9 * q ** 1.08)

        T_c_f = 1. / (1. + (ks/5.4) ** 4)
        f = lambda a, b : a / (a + b*q**2)
        T_c = T_c_f * f(T_c_ln_beta, T_c_C_noalpha) + (1-T_c_f) * f(T_c_ln_beta, T_c_C_alpha)

        s_tilde = self.sound_horizon * (1 + (self.beta_node/ks)**3) ** (-1./3.)
        ks_tilde = k*s_tilde

        T_b_T0 = f(T_c_ln_nobeta, T_c_C_noalpha)
        T_b_1 = T_b_T0 / (1 + (ks/5.2)**2 )
        T_b_2 = self.alpha_b / (1 + (self.beta_b/ks)**3 ) * numpy.exp(-(k/self.k_silk) ** 1.4)
        T_b = numpy.sinc(ks_tilde/numpy.pi) * (T_b_1 + T_b_2)

        T = numpy.ones(valid.shape)
        T[valid] = self.f_baryon*T_b + (1-self.f_baryon)*T_c;
        return T * self.cosmo.scale_independent_growth_factor(self.redshift)

class NoWiggleEisensteinHu(object):
    """
    Linear power spectrum using the Eisenstein & Hu (1998) fitting formula
    without BAO wiggles.

    Parameters
    ----------
    cosmo : :class:`Cosmology`
        the cosmology instance
    redshift : float
        the redshift of the power spectrum

    References
    ----------
    Eisenstein & Hu, "Baryonic Features in the Matter Transfer Function", 1998
    """
    def __init__(self, cosmo, redshift):
        self.cosmo = cosmo
        self.redshift = redshift

        self.Obh2      = cosmo.Omega0_b * cosmo.h ** 2
        self.Omh2      = cosmo.Omega0_m * cosmo.h ** 2
        self.f_baryon  = cosmo.Omega0_b / cosmo.Omega0_m
        self.theta_cmb = cosmo.Tcmb0 / 2.7

        # wavenumber of equality
        self.k_eq = 0.0746 * self.Omh2 * self.theta_cmb ** (-2) # units of 1/Mpc

        self.sound_horizon = cosmo.h * 44.5 * numpy.log(9.83/self.Omh2) / \
                            numpy.sqrt(1 + 10 * self.Obh2** 0.75) # in Mpc/h
        self.alpha_gamma = 1 - 0.328 * numpy.log(431*self.Omh2) * self.f_baryon + \
                            0.38* numpy.log(22.3*self.Omh2) * self.f_baryon ** 2


    def __call__(self, k):
        r"""
        Return the Eisenstein-Hu transfer function without BAO wiggles.

        This is normalized to unity as :math:`k \rightarrow 0` at :math:`z=0`.
        The redshift scaling is provided by the
        :func:`~Cosmology.scale_independent_growth_factor` function.

        Parameters
        ---------
        k : float, array_like
            the wavenumbers in units of :math:`h \mathrm{Mpc}^{-1}`

        Returns
        -------
        Tk : float, array_like
            the transfer function evaluated at ``k``, normalized to unity on
            large scales
        """
        if numpy.isscalar(k) and k == 0.:
            return 1.0

        # only compute k > 0 modes
        k = numpy.asarray(k)
        valid = k > 0.

        k = k[valid] * self.cosmo.h # in 1/Mpc now
        ks = k * self.sound_horizon / self.cosmo.h
        q = k / (13.41*self.k_eq)

        gamma_eff = self.Omh2 * (self.alpha_gamma + (1 - self.alpha_gamma) / (1 + (0.43*ks) ** 4))
        q_eff = q * self.Omh2 / gamma_eff
        L0 = numpy.log(2*numpy.e + 1.8 * q_eff)
        C0 = 14.2 + 731.0 / (1 + 62.5 * q_eff)

        T = numpy.ones(valid.shape)
        T[valid] = L0 / (L0 + C0 * q_eff**2)
        return T * self.cosmo.scale_independent_growth_factor(self.redshift)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from nbodykit.meshtools import SlabIterator

from pmesh.pm import ParticleMesh, RealField, ComplexField
import pytest
from numpy.testing import assert_array_equal

# debug logging
setup_logging("debug")

@MPITest([1])
def test_wrong_ndim(comm):

    numpy.random.seed(42)

    pm = ParticleMesh(BoxSize=8.0, Nmesh=[8,8], comm=comm, dtype='f8')
    rfield = RealField(pm)
    data = numpy.random.random(size=rfield.shape)
    rfield[...] = data[:]

    # SlabIterator only works for 2D or 3D coordinate meshes
    with pytest.raises(NotImplementedError):
        for slab in SlabIterator([rfield.x[0]], axis=0, symmetry_axis=None):
            pass

@MPITest([1])
def test_wrong_coords_shape(comm):

    numpy.random.seed(42)

    pm = ParticleMesh(BoxSize=8.0, Nmesh=[8, 8], comm=comm, dtype='f8')
    rfield = RealField(pm)
    data = numpy.random.random(size=rfield.shape)
    rfield[...] = data[:]

    # coords arrays should not be squeezed
    x = [numpy.squeeze(xx) for xx in rfield.x]

    with pytest.raises(ValueError):
        for slab in SlabIterator(x, axis=0, symmetry_axis=None):
            pass


@MPITest([1, 4])
def test_2d_slab(comm):

    numpy.random.seed(42)

    pm = ParticleMesh(BoxSize=8.0, Nmesh=[8, 8], comm=comm, dtype='f8')
    rfield = RealField(pm)
    data = numpy.random.random(size=rfield.shape)
    rfield[...] = data[:]

    x = rfield.x
    for i, slab in enumerate(SlabIterator(x, axis=0, symmetry_axis=None)):
        assert slab.__str__() == slab.__repr__()
        assert slab.shape == (8,)
        assert_array_equal(slab.hermitian_weights, numpy.ones(slab.shape))
        assert_array_equal(rfield[slab.index],  data[i])

@MPITest([1, 4])
def test_hermitian_weights(comm):

    numpy.random.seed(42)

    pm = ParticleMesh(BoxSize=8.0, Nmesh=[8, 8, 8], comm=comm, dtype='f8')
    cfield = ComplexField(pm)
    data = numpy.random.random(size=cfield.shape)
    data = data[:] + 1j*data[:]

    cfield[...] = data[:]
    x = cfield.x

    # iterate over symmetry axis
    for i, slab in enumerate(SlabIterator(x, axis=2, symmetry_axis=2)):

        # nonsingular weights give indices of positive frequencies
        nonsig = slab.nonsingular
        weights = slab.hermitian_weights

        # weights == 2 when iterating frequency is positive
        if numpy.float(slab.coords(2)) > 0.:
            assert weights > 1
            assert numpy.all(nonsig == True)
        else:
            assert weights == 1.0
            assert numpy.all(nonsig == False)



from runtests.mpi import MPITest
from nbodykit.source.catalog import SubVolumesCatalog
from numpy.testing import assert_allclose, assert_array_equal
from nbodykit import setup_logging
import pytest
from nbodykit.lab import *

setup_logging("debug")

@MPITest([1, 4])
def test_decomposed(comm):
    cosmo = cosmology.Planck15

    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=1e-5, BoxSize=128., Nmesh=8, seed=42, comm=comm)

    decomposed = source.to_subvolumes()

    assert decomposed.comm == source.comm
    assert 'Position' in decomposed
    assert 'Velocity' in decomposed

    assert len(decomposed['Position'].compute()) == decomposed.size
    assert len(decomposed['Velocity'].compute()) == decomposed.size

    decomposed = source.to_subvolumes(columns=['Position'])
    assert decomposed.comm == source.comm

    assert 'Position' in decomposed
    assert 'Velocity' not in decomposed

    assert len(decomposed['Position'].compute()) == decomposed.size
